# coding: utf-8

import os
from glob import glob

import numpy
import pytest

from tomoscan.esrf.identifier.folderidentifier import BaseFolderIdentifierMixIn
from tomoscan.esrf.volume.edfvolume import EDFVolume
from tomoscan.esrf.volume.jp2kvolume import JP2KVolume, has_glymur, has_minimal_openjpeg
from tomoscan.esrf.volume.mock import create_volume
from tomoscan.esrf.volume.tiffvolume import TIFFVolume, has_tifffile
from tomoscan.factory import Factory
from tomoscan.identifier import VolumeIdentifier
from tomoscan.volumebase import VolumeBase

_data = create_volume(
    frame_dims=(100, 100), z_size=11
)  # z_size need to be at least 10 to check loading from file name works
for i in range(len(_data)):
    _data[i] += 1
_data = _data.astype(numpy.uint16)


_metadata = {
    "nabu_config": {
        "dataset": {
            "location": "toto.hdf5",
            "entry": "entry0000",
        },
        "phase": {
            "method": "None",
            "padding_type": "edge",
        },
    },
    "processing_option": {
        "build_sino": {
            "axis_correction": "None",
            "enable_halftomo": True,
        },
        "flatfield": {
            "binning": "11",
            "do_flat_distortion": False,
        },
    },
}


volume_constructors = [
    EDFVolume,
]
if has_tifffile:
    volume_constructors.append(TIFFVolume)
if has_glymur and has_minimal_openjpeg:
    volume_constructors.append(JP2KVolume)


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_create_volume_from_folder(tmp_path, volume_constructor):
    """Test a volume can be weel defined from it path"""
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = str(acquisition_dir / "volume")
    os.makedirs(volume_dir)
    volume = volume_constructor(folder=volume_dir, data=_data, metadata=_metadata)
    assert (
        len(
            glob(
                os.path.join(
                    volume_dir, f"*.{volume_constructor.DEFAULT_DATA_EXTENSION}"
                )
            )
        )
        == 0
    )
    assert (
        len(
            glob(
                os.path.join(
                    volume_dir, f"*.{volume_constructor.DEFAULT_METADATA_EXTENSION}"
                )
            )
        )
        == 0
    )
    assert tuple(volume.browse_data_files()) == tuple()
    assert tuple(volume.browse_data_urls()) == tuple()
    assert tuple(volume.browse_metadata_files()) == tuple()
    volume.save()
    assert (
        len(
            glob(
                os.path.join(
                    volume_dir, f"*.{volume_constructor.DEFAULT_DATA_EXTENSION}"
                )
            )
        )
        == _data.shape[0]
    )
    assert (
        len(
            glob(
                os.path.join(
                    volume_dir, f"*.{volume_constructor.DEFAULT_METADATA_EXTENSION}"
                )
            )
        )
        == 1
    )

    # check overwrite parameter
    volume.data = numpy.random.random((45, 45, 45))
    with pytest.raises(OSError):
        volume.save()

    volume.overwrite = True
    if isinstance(volume, JP2KVolume):
        volume.rescale_data = False
    volume.save()
    assert volume.get_volume_shape() == (45, 45, 45)
    volume.data = _data
    volume.save()

    # check load data and metadata
    volume.clear_cache()
    assert volume.get_min_max_values() == (_data.min(), _data.max())
    volume.load()
    numpy.testing.assert_array_almost_equal(_data, volume.data)
    assert _metadata == volume.metadata

    # test get_slice function
    numpy.testing.assert_array_equal(
        volume.get_slice(index=2, axis=0),
        _data[2],
    )
    numpy.testing.assert_array_equal(
        volume.get_slice(index=2, axis=1),
        _data[:, 2, :],
    )
    numpy.testing.assert_array_equal(
        volume.get_slice(index=2, axis=2),
        _data[:, :, 2],
    )
    # test deprecated API
    numpy.testing.assert_array_equal(
        volume.get_slice(index=2, axis=0),
        volume.get_slice(xy=2),
    )
    numpy.testing.assert_array_equal(
        volume.get_slice(index=2, axis=1),
        volume.get_slice(xz=2),
    )
    numpy.testing.assert_array_equal(
        volume.get_slice(index=2, axis=2),
        volume.get_slice(yz=2),
    )

    assert len(list(volume.browse_data_files())) == _data.shape[0]
    assert len(list(volume.browse_data_urls())) == _data.shape[0]
    assert len(list(volume.browse_metadata_files())) == 1

    # test browsing frames
    for l_frame, o_frame in zip(volume.browse_slices(), _data):
        numpy.testing.assert_allclose(l_frame, o_frame)

    # test get_min_max values
    volume.data = None
    assert volume.get_min_max() == (_data.min(), _data.max())
    volume.load_data()
    assert volume.get_min_max() == (_data.min(), _data.max())

    if isinstance(volume, JP2KVolume):
        volume.data_extension == "jp2k"
    elif isinstance(volume, EDFVolume):
        volume.data_extension == "edf"
    elif isinstance(volume, TIFFVolume):
        volume.data_extension == "tiff"
    else:
        raise NotImplementedError
    assert volume.metadata_extension == "txt"


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_get_start_index(tmp_path, volume_constructor):
    """Test setting start_index parameter"""
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = str(acquisition_dir / "volume")
    os.makedirs(volume_dir)

    # use data which has a different value for each frame
    data = _data.copy()
    data += numpy.arange(data.shape[0], dtype=_data.dtype).reshape((-1, 1, 1))
    #

    start_index_list = [0, 7]  # one default and one non-default
    for start_index in start_index_list:
        volume_basename = f"test_startindex{start_index}"

        volume = volume_constructor(
            folder=volume_dir, volume_basename=volume_basename, start_index=start_index
        )
        assert volume.get_first_file() is None
        volume.data = _data
        volume.save()

        expected_first_file_name = os.path.join(
            volume_dir,
            "%s_%06d.%s"
            % (volume_basename, start_index, volume_constructor.DEFAULT_DATA_EXTENSION),
        )
        assert volume.get_first_file() == expected_first_file_name
        assert volume.get_file_slice_index(volume.get_first_file()) == start_index

        # Now create another XXVolume object using the same basename, while there are existing files
        volume2 = volume_constructor(
            folder=volume_dir, volume_basename=volume_basename, start_index=start_index
        )
        assert volume2.get_first_file() == expected_first_file_name
        assert (
            volume2.get_file_slice_index(volume.get_first_file()) == start_index
        )  # This used to be False before this patch


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_reset_slice_index(tmp_path, volume_constructor):
    """
    Test a case that might happen during volume casting: existing files have start_index > 0,
    so we should re-define start_index for the new Volume object before writing
    """
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = str(acquisition_dir / "volume")
    os.makedirs(volume_dir)
    volume_dir2 = str(acquisition_dir / "volume2")
    os.makedirs(volume_dir2)

    # use data which has a different value for each frame
    data = _data.copy()
    data += numpy.arange(data.shape[0], dtype=_data.dtype).reshape((-1, 1, 1))
    #

    # Step 1: a Volume is created during reconstruction, with start_z > 0
    volume = volume_constructor(
        folder=volume_dir, volume_basename="test_reset_slice_index", start_index=500
    )
    if isinstance(volume, JP2KVolume):
        volume.rescale_data = False
    volume.data = data
    volume.save()

    # Step 2: after reconstruction, volume casting is triggered.
    # The "input_volume" object has to detect the actual start_index
    input_volume = volume_constructor(
        folder=volume_dir, volume_basename="test_reset_slice_index"
    )
    output_volume = volume_constructor(
        folder=volume_dir2, volume_basename="test_reset_slice_index"
    )
    if isinstance(output_volume, JP2KVolume):
        output_volume.rescale_data = False
    input_vol_start_idx = input_volume.get_file_slice_index(
        input_volume.get_first_file()
    )
    assert input_vol_start_idx == 500
    output_volume.start_index = input_vol_start_idx
    output_volume.data = data + 10
    output_volume.save()

    # Check !
    vol_check = volume_constructor(
        folder=volume_dir2, volume_basename="test_reset_slice_index", start_index=500
    )
    assert numpy.allclose(vol_check.load_data(), volume.load_data() + 10)


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_data_file_saver_generator(tmp_path, volume_constructor):
    """
    data_file_saver_generator (dumping frame by frame a volume to disk)
    """
    volume_dir = str(tmp_path / "volume")
    os.makedirs(volume_dir)
    volume = volume_constructor(folder=volume_dir)
    if isinstance(volume, JP2KVolume):
        volume.rescale_data = False
    for slice_, slice_saver in zip(
        _data,
        volume.data_file_saver_generator(
            n_frames=_data.shape[0], data_url=volume.data_url, overwrite=False
        ),
    ):
        slice_saver[:] = slice_
    assert volume.data is None
    numpy.testing.assert_array_equal(
        volume.load_data(),
        _data,
    )


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_decreasing_index(tmp_path, volume_constructor):
    """
    Test the 'index_step' parameter
    """
    volume_dir = str(tmp_path / "volume")
    os.makedirs(volume_dir)
    volume = volume_constructor(folder=volume_dir, start_index=200)
    volume.write_in_descending_order = True
    if isinstance(volume, JP2KVolume):
        volume.rescale_data = False
    data = numpy.arange(32 * 33 * 34, dtype="f").reshape(
        (32, 33, 34)
    )  # put a not-too-small number for jpeg2000
    volume.data = data
    volume.save()
    # Check that the indices are correct (this will not check the write index order)
    files_list = sorted(
        glob(os.path.join(volume_dir, f"*.{volume.DEFAULT_DATA_EXTENSION}"))
    )
    indices = [
        int(os.path.splitext(os.path.basename(fname))[0].split("_")[-1])
        for fname in files_list
    ]
    assert sorted(indices, reverse=True) == list(
        range(volume.start_index, volume.start_index - data.shape[0], -1)
    )

    # Check that the files order was reversed at write time
    volume_reader = volume_constructor(folder=volume_dir)
    data_read = volume_reader.load_data()
    assert numpy.allclose(data_read, data[::-1, ...])


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_several_writer(tmp_path, volume_constructor):
    """
    Test writing a volume from several instance of VolumeSingleFrameBase and reading the entire volume back
    """
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = acquisition_dir / "volume"
    os.makedirs(volume_dir)

    volume_1 = volume_constructor(
        folder=volume_dir,
        data=_data[:5],
        metadata=_metadata,
    )
    volume_2 = volume_constructor(folder=volume_dir, data=_data[5:], start_index=5)
    if isinstance(volume_1, JP2KVolume):
        volume_1.rescale_data = (
            False  # keep coherence between all the volumes. Simplify test
        )
        volume_2.rescale_data = False
    volume_1.save()
    volume_2.skip_existing_data_files_removal = True
    volume_2.save()

    full_volume = volume_constructor(folder=volume_dir)
    full_volume.load()
    numpy.testing.assert_array_almost_equal(full_volume.data, _data)
    assert full_volume.metadata == _metadata
    full_volume.data = None
    assert full_volume.get_volume_shape() == _data.shape


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_volume_identifier(tmp_path, volume_constructor):
    """
    Insure each type of volume can provide an identifier and recover the data store from it
    """
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = acquisition_dir / "volume"
    os.makedirs(volume_dir)

    volume = volume_constructor(
        folder=volume_dir,
        data=_data,
        metadata=_metadata,
    )
    if isinstance(volume, JP2KVolume):
        volume.rescale_data = (
            False  # keep coherence between all the volumes. Simplify test
        )
    volume.save()
    identifier = volume.get_identifier()
    assert isinstance(identifier, VolumeIdentifier)

    del volume
    volume_loaded = volume_constructor.from_identifier(identifier)
    volume_loaded.load()
    numpy.testing.assert_array_almost_equal(volume_loaded.data, _data)
    assert volume_loaded.metadata == _metadata

    # check API Identifier.to() and identifier.from_str()
    identifier_from_str = identifier.to_str()
    duplicate_id = identifier.from_str(identifier_from_str)
    assert duplicate_id == identifier
    assert identifier_from_str == identifier.to_str()
    assert identifier != object()
    assert identifier != "toto"
    assert isinstance(identifier.short_description(), str)

    # check it can be reconstructed from the Factory
    assert isinstance(identifier, VolumeIdentifier)
    tomo_obj_from_str = Factory.create_tomo_object_from_identifier(
        identifier=identifier
    )
    assert isinstance(tomo_obj_from_str, type(volume_loaded))
    tomo_obj_from_identifier = Factory.create_tomo_object_from_identifier(
        identifier=identifier.to_str()
    )
    assert isinstance(tomo_obj_from_identifier, type(volume_loaded))

    # test hash
    hash(tomo_obj_from_str)
    hash(tomo_obj_from_str.get_identifier())


def test_folder_mix_in():
    """simple test of the BaseFolderIdentifierMixIn class"""

    class FolderIdentifierTest(BaseFolderIdentifierMixIn, VolumeIdentifier):
        pass

    class VolumeTest(VolumeBase):
        pass

    obj = FolderIdentifierTest(object=VolumeTest, folder="toto", tomo_type="Volume")
    with pytest.raises(NotImplementedError):
        obj.scheme


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_volume_with_prefix(tmp_path, volume_constructor):
    """
    Test writing and reading volume with a file_prefix and using the identifier
    """
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = acquisition_dir / "volume"
    os.makedirs(volume_dir)
    file_prefix = "test"

    volume_1 = volume_constructor(
        folder=volume_dir,
        data=_data[:5],
        metadata=_metadata,
        volume_basename=file_prefix,
    )
    if isinstance(volume_1, JP2KVolume):
        volume_1.rescale_data = (
            False  # keep coherence between all the volumes. Simplify test
        )
    volume_1.save()

    full_volume = volume_constructor(folder=volume_dir, volume_basename=file_prefix)
    full_volume.load()
    numpy.testing.assert_array_almost_equal(full_volume.data, _data[:5])
    assert full_volume.metadata == _metadata

    obj_recreated = Factory.create_tomo_object_from_identifier(
        full_volume.get_identifier()
    )
    assert isinstance(obj_recreated, volume_constructor)
    obj_recreated.load()

    numpy.testing.assert_array_equal(
        obj_recreated.data,
        volume_1.data,
    )

    numpy.testing.assert_array_equal(
        obj_recreated.metadata,
        volume_1.metadata,
    )


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_example(volume_constructor):
    """test static function 'example'"""
    assert isinstance(volume_constructor.example_defined_from_str_identifier(), str)


def test_backward_compatibility_z_index_4(tmp_path):
    """
    make sure we are still able to read old volumes with the new class
    """
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = str(acquisition_dir / "volume")
    os.makedirs(volume_dir)

    class LegacyEDFVolume(EDFVolume):
        DEFAULT_DATA_PATH_PATTERN = "{volume_basename}_{index_zfill4}.{data_extension}"

    legacy_volume = LegacyEDFVolume(
        folder=volume_dir,
        data=_data,
    )
    legacy_volume.save()
    first_frame_path = os.path.join(volume_dir, "volume_0000.edf")
    assert os.path.exists(first_frame_path)

    # test loading with the legacy class
    legacy_volume.clear_cache()

    legacy_volume.load()
    numpy.testing.assert_array_almost_equal(_data, legacy_volume.data)

    # test loading with the updated EDFVolume class
    volume = EDFVolume(
        folder=volume_dir,
    )
    volume.load()
    numpy.testing.assert_array_almost_equal(_data, legacy_volume.data)


@pytest.mark.parametrize("save_volume_to_file", (True, False))
@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_get_slices(tmp_path, volume_constructor, save_volume_to_file):
    """test the volume 'get_slices' function"""
    test_dir = tmp_path / "test_get_slices"
    test_dir.mkdir()

    volume_dir = os.path.join(test_dir, "volume")
    volume_shape = 20, 100, 50
    volume_n_elemts = volume_shape[0] * volume_shape[1] * volume_shape[2]
    data = numpy.linspace(
        0,
        volume_n_elemts,
        num=volume_n_elemts,
        endpoint=False,
    ).reshape(volume_shape)
    volume = volume_constructor(folder=volume_dir, data=data, metadata={})

    if save_volume_to_file:
        volume.save()
        volume.clear_cache()

    # test raising error

    with pytest.raises(TypeError):
        volume.get_slices(
            "toto",
        )

    with pytest.raises(TypeError):
        volume.get_slices(
            (0, 10),
        )

    # test retrieving results
    slices_axis_0 = volume.get_slices(((0, 2), (0, 5)))
    slices_axis_1 = volume.get_slices(((1, 0), (1, 6)))
    slices_axis_2 = volume.get_slices(((2, 8), (2, 9)))
    slices_several_axis = volume.get_slices(((0, 2), (1, 0), (2, 9)))

    if volume_constructor is JP2KVolume:
        # JP2K is modifying original slice values. So we cannot compare them with
        # the original one.
        return

    numpy.testing.assert_array_almost_equal(
        slices_axis_0[(0, 2)],
        data[2],
    )

    numpy.testing.assert_array_almost_equal(
        slices_axis_1[(1, 6)],
        data[:, 6, :],
    )

    numpy.testing.assert_array_almost_equal(
        slices_axis_2[(2, 9)],
        data[:, :, 9],
    )

    numpy.testing.assert_array_almost_equal(
        slices_several_axis[(0, 2)],
        slices_axis_0[(0, 2)],
    )

    numpy.testing.assert_array_almost_equal(
        slices_several_axis[(1, 0)],
        slices_axis_1[(1, 0)],
    )

    numpy.testing.assert_array_almost_equal(
        slices_several_axis[(2, 9)],
        slices_axis_2[(2, 9)],
    )


@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_load_chunk(tmp_path, volume_constructor):
    volume_dir = str(tmp_path / "volume")
    os.makedirs(volume_dir)
    volume = volume_constructor(folder=volume_dir)

    if isinstance(volume, JP2KVolume):  # bis repetita (non) placent
        volume.rescale_data = False
    volume.data = numpy.arange(32 * 33 * 34, dtype="f").reshape(
        (32, 33, 34)
    )  # put a not-too-small number for jpeg2000
    volume.save_data()

    volume_reader = volume_constructor(folder=volume_dir)
    volume_reader.load_data()
    data_all = volume_reader.data

    chunks_to_test = [
        (slice(1, 3), slice(None), slice(None)),
        (slice(-7, None), slice(None), slice(None)),
        (slice(None), slice(0, 3), slice(None)),
        (slice(None), slice(None), slice(5, 9)),
        (slice(None), slice(None), slice(None)),
    ]
    for chunk in chunks_to_test:
        data = volume.load_chunk(chunk)
        assert numpy.allclose(data, data_all[chunk])



# coding: utf-8
"""Module containing validators"""

from __future__ import annotations

import logging
import weakref

import numpy
from silx.io.utils import get_data

from tomoscan.esrf.scan.utils import dataset_has_broken_vds, get_compacted_dataslices
from tomoscan.scanbase import TomoScanBase

_logger = logging.getLogger(__name__)

__all__ = [
    "ValidatorBase",
    "DarkEntryValidator",
    "DarkDatasetValidator",
    "FlatEntryValidator",
    "FlatDatasetValidator",
    "ProjectionEntryValidator",
    "ProjectionDatasetValidator",
    "EnergyValidator",
    "DistanceValidator",
    "PixelValidator",
    "BasicScanValidator",
    "ReconstructionValidator",
    "is_valid_for_reconstruction",
]

_VALIDATOR_NAME_TXT_AJUST = 15

_LOCATION_TXT_AJUST = 40

_SCAN_NAME_TXT_AJUST = 30

_BOMB_UCODE = "\U0001f4a3"

_EXPLOSION_UCODE = "\U0001f4a5"

_THUMB_UP_UCODE = "\U0001f44d"

_OK_UCODE = "\U0001f44c"


class ValidatorBase:
    """Base validator class"""

    def is_valid(self) -> bool:
        raise NotImplementedError("Base class")

    def run(self) -> bool:
        raise NotImplementedError("Base class")

    def clear(self) -> None:
        raise NotImplementedError("Base class")


class _ScanParamValidator(ValidatorBase):
    def __init__(self, scan: TomoScanBase, name: str, location: str | None):
        if not isinstance(scan, TomoScanBase):
            raise TypeError(f"{scan} is expected to be an instance of {TomoScanBase}")
        self._scan = weakref.ref(scan)
        self.__name = name
        self.__location = location
        self._valid = None

    @property
    def name(self):
        return self.__name

    def __str__(self):
        return self.info()

    def info(self, with_scan=True):
        info = [
            self.name.ljust(_VALIDATOR_NAME_TXT_AJUST) + ":",
            "VALID".ljust(7) if self.is_valid() else "INVALID".ljust(7),
        ]
        if with_scan:
            info.insert(
                0,
                str(self.scan).ljust(_SCAN_NAME_TXT_AJUST) + " - ",
            )
        if not self.is_valid():
            info.append(
                f"Expected location: {self.__location}".ljust(_LOCATION_TXT_AJUST)
            )
        return " ".join(info)

    def _run(self):
        """Function to overwrite to compute the validity condition"""
        raise NotImplementedError("Base class")

    @property
    def scan(self) -> TomoScanBase | None:
        if self._scan and self._scan():
            return self._scan()
        else:
            return None

    def is_valid(self) -> bool:
        if self._valid is None:
            self._valid = self.run()
        return self._valid

    def clear(self):
        self._valid = None

    def run(self) -> bool | None:
        """
        Return None if unable to find if valid or not. Otherwise a boolean
        """
        if self.scan is None:
            self._valid = None
            return None
        else:
            return self._run()


class DarkEntryValidator(_ScanParamValidator):
    """
    Check darks are present and valid
    """

    def __init__(self, scan):
        super().__init__(
            scan=scan,
            name="dark(s)",
            location=scan.get_dark_expected_location(),
        )

    def _run(self) -> None:
        return self.scan.darks is not None and len(self.scan.darks) > 0


class _VdsAndValuesValidatorMixIn:
    def __init__(self, check_values, check_vds):
        self._check_values = check_values
        self._check_vds = check_vds
        self._has_data = None
        self._vds_ok = None
        self._no_nan = None

    @property
    def is_valid(self):
        raise NotImplementedError("Base class")

    @property
    def name(self):
        raise NotImplementedError("Base class")

    @property
    def scan(self):
        raise NotImplementedError("Base class")

    @property
    def location(self):
        raise NotImplementedError("Base class")

    @property
    def check_values(self):
        return self._check_values

    @property
    def check_vds(self):
        return self._check_vds

    def check_urls(self, urls: dict):
        if urls is None:
            return True

        _, compacted_urls = get_compacted_dataslices(urls, return_url_set=True)

        if self.check_vds:
            # compact urls to speed up
            for _, url in compacted_urls.items():
                if dataset_has_broken_vds(url=url):
                    self._vds_ok = False
                    return False
            else:
                self._vds_ok = True

        if self.check_values:
            self._no_nan = True
            for _, url in compacted_urls.items():
                data = get_data(url)
                self._no_nan = self._no_nan and not numpy.isnan(data).any()
            return self._no_nan
        return True

    def clear(self):
        self._has_data = None
        self._vds_ok = None
        self._no_nan = None

    def info(self, with_scan=True):
        text = "VALID".ljust(7) if self.is_valid() else "INVALID".ljust(7)
        if not self._has_data:
            text = " - ".join(
                (text, f"Unable to find data. Expected location: {self.location}")
            )
        elif self.check_vds and not self._vds_ok:
            text = " - ".join((text, "At least one dataset seems to have broken link"))
        elif self.check_values and not self._no_nan:
            text = " - ".join(
                (text, "At least one dataset seems to contains `nan` value")
            )

        text = [
            f"{self.name}".ljust(_VALIDATOR_NAME_TXT_AJUST) + ":",
            text,
        ]
        if with_scan:
            text.insert(0, f"{str(self.scan)}".ljust(_SCAN_NAME_TXT_AJUST) + ",")

        return " ".join(text)


class DarkDatasetValidator(DarkEntryValidator, _VdsAndValuesValidatorMixIn):
    """Check entries exists and values are valid"""

    def __init__(self, scan, check_vds, check_values):
        DarkEntryValidator.__init__(self, scan=scan)
        _VdsAndValuesValidatorMixIn.__init__(
            self, check_vds=check_vds, check_values=check_values
        )

    def _run(self) -> bool:
        # check darks exists
        self._has_data = DarkEntryValidator._run(self)
        if self._has_data is False:
            return False

        return _VdsAndValuesValidatorMixIn.check_urls(self, self.scan.darks)

    def info(self, with_scan=True):
        return _VdsAndValuesValidatorMixIn.info(self, with_scan)


class FlatEntryValidator(_ScanParamValidator):
    """
    Check flats are present and valid
    """

    def __init__(self, scan):
        super().__init__(
            scan=scan, name="flat(s)", location=scan.get_flat_expected_location()
        )

    def _run(self) -> bool | None:
        return self.scan.flats is not None and len(self.scan.flats) > 0


class FlatDatasetValidator(FlatEntryValidator, _VdsAndValuesValidatorMixIn):
    """Check entries exists and values are valid"""

    def __init__(self, scan, check_vds, check_values):
        FlatEntryValidator.__init__(self, scan=scan)
        _VdsAndValuesValidatorMixIn.__init__(
            self, check_vds=check_vds, check_values=check_values
        )

    def _run(self) -> bool:
        # check darks exists
        self._has_data = FlatEntryValidator._run(self)
        if self._has_data is False:
            return False

        return _VdsAndValuesValidatorMixIn.check_urls(self, self.scan.flats)

    def info(self, with_scan=True):
        return _VdsAndValuesValidatorMixIn.info(self, with_scan)


class ProjectionEntryValidator(_ScanParamValidator):
    """
    Check at projections are present and seems coherent with what is expected
    """

    def __init__(self, scan):
        super().__init__(
            scan=scan,
            name="projection(s)",
            location=scan.get_projection_expected_location(),
        )

    def _run(self) -> bool | None:
        if self.scan.projections is None:
            return False
        elif self.scan.tomo_n is not None:
            return len(self.scan.projections) == self.scan.tomo_n
        else:
            return len(self.scan.projections) > 0


class ProjectionDatasetValidator(ProjectionEntryValidator, _VdsAndValuesValidatorMixIn):
    """Check projections frames exists and values seems valid"""

    def __init__(self, scan, check_vds, check_values):
        ProjectionEntryValidator.__init__(self, scan=scan)
        _VdsAndValuesValidatorMixIn.__init__(
            self, check_vds=check_vds, check_values=check_values
        )

    def _run(self) -> bool:
        # check darks exists
        self._has_data = ProjectionEntryValidator._run(self)
        if self._has_data is False:
            return False

        return _VdsAndValuesValidatorMixIn.check_urls(self, self.scan.projections)

    def info(self, with_scan=True):
        return _VdsAndValuesValidatorMixIn.info(self, with_scan)


class EnergyValidator(_ScanParamValidator):
    """Check energy can be read and is not 0"""

    def __init__(self, scan):
        super().__init__(
            scan=scan,
            name="energy",
            location=scan.get_energy_expected_location(),
        )

    def _run(self) -> bool | None:
        return self.scan.energy not in (None, 0)


class DistanceValidator(_ScanParamValidator):
    """Check distance can be read and is not 0"""

    def __init__(self, scan):
        super().__init__(
            scan=scan,
            name="distance",
            location=scan.get_sample_detector_distance_expected_location(),
        )

    def _run(self) -> bool | None:
        return self.scan.sample_detector_distance not in (None, 0)


class PixelValidator(_ScanParamValidator):
    """Check pixel size can be read and is / are not 0"""

    def __init__(self, scan):
        super().__init__(
            scan=scan,
            name="pixel size",
            location=scan.get_pixel_size_expected_location(),
        )

    def _run(self) -> bool | None:
        from tomoscan.esrf.scan.nxtomoscan import NXtomoScan

        if isinstance(self.scan, NXtomoScan):
            return (self.scan.sample_x_pixel_size not in (None, 0)) and (
                self.scan.sample_y_pixel_size not in (None, 0)
            )
        else:
            return self.scan.pixel_size not in (None, 0)


class _ValidatorGroupMixIn:
    """
    Represents a group of validators.
    Define a `checkup` function to display a resume of valid and invalid tasks
    """

    def __init__(self):
        self._validators = []

    def checkup(self, only_issues=False) -> str:
        """
        compute a short text with:
         * if only_issues is False: all information checked and the status of the information
         * if only_issues is true: all mandatory information missing
        """

        def _is_invalid(validator):
            return not validator.is_valid()

        validators_with_issues = tuple(filter(_is_invalid, self._validators))

        def get_first_chars(validator):
            if validator.is_valid():
                return "+"
            else:
                return "-"

        if only_issues:
            if len(validators_with_issues) == 0:
                text = self.get_text_no_issue() + "\n"
            else:
                text = [
                    f"   {get_first_chars(validator)} {validator.info(with_scan=False)}"
                    for validator in validators_with_issues
                ]
                text.insert(0, self.get_text_issue(len(validators_with_issues)))
                text.append(" ")
                text = "\n".join(text)
        else:
            text = [
                f"   {get_first_chars(validator)} {validator.info(with_scan=False)}"
                for validator in self._validators
            ]
            if len(validators_with_issues) == 0:
                text.insert(0, self.get_text_no_issue())
            else:
                text.insert(0, self.get_text_issue(len(validators_with_issues)))
            text.append(" ")
            text = "\n".join(text)

        return text

    def is_valid(self) -> bool:
        valid = True
        for validator in self._validators:
            assert isinstance(
                validator, ValidatorBase
            ), "validators should be instances of ValidatorBase"
            valid = valid + validator.is_valid()
        return valid

    def _run(self) -> bool | None:
        run_ok = True
        for validator in self._validators:
            run_ok = run_ok and validator.run()
        return run_ok

    def clear(self) -> None:
        [validator.clear() for validator in self._validators]

    def get_text_no_issue(self) -> str:
        raise NotImplementedError("Base class")

    def get_text_issue(self, n_issue) -> str:
        raise NotImplementedError("Base class")


class BasicScanValidator(_ValidatorGroupMixIn, ValidatorBase):
    """Check that a scan has some basic parameters as dark, flat..."""

    def __init__(
        self, scan, check_vds=True, check_dark=True, check_flat=True, check_values=False
    ):
        super(BasicScanValidator, self).__init__()
        if not isinstance(scan, TomoScanBase):
            raise TypeError(f"{scan} is expected to be an instance of {TomoScanBase}")
        self._scan = scan

        self._validators.append(
            ProjectionDatasetValidator(
                scan=scan, check_values=check_values, check_vds=check_vds
            )
        )

        if check_dark:
            self._validators.append(
                DarkDatasetValidator(
                    scan=scan, check_values=check_values, check_vds=check_vds
                )
            )
        if check_flat:
            self._validators.append(
                FlatDatasetValidator(
                    scan=scan, check_values=check_values, check_vds=check_vds
                )
            )

    @property
    def scan(self):
        return self._scan

    def get_text_no_issue(self) -> str:
        header = f"{_OK_UCODE}{_THUMB_UP_UCODE}{_OK_UCODE}"
        return f"{header}\n No issue found from {self.scan}."

    def get_text_issue(self, n_issue) -> str:
        header = f"{_EXPLOSION_UCODE}{_BOMB_UCODE}{_EXPLOSION_UCODE}"
        return f"{header}\n {n_issue} issues found from {self.scan}"


class ReconstructionValidator(BasicScanValidator):
    """
    Check that a dataset/scan has enough valid parameters to be reconstructed
    by a software like nabu
    """

    def __init__(
        self,
        scan: TomoScanBase,
        check_phase_retrieval=True,
        check_values=False,
        check_vds=True,
        check_dark=True,
        check_flat=True,
    ):
        super().__init__(
            scan=scan,
            check_dark=check_dark,
            check_flat=check_flat,
            check_values=check_values,
            check_vds=check_vds,
        )
        self._need_phase_retrieval = check_phase_retrieval
        if self.check_phase_retrieval:
            self._validators.append(DistanceValidator(scan=scan))
            self._validators.append(EnergyValidator(scan=scan))
            self._validators.append(PixelValidator(scan=scan))

    @property
    def check_phase_retrieval(self):
        return self._need_phase_retrieval

    @check_phase_retrieval.setter
    def check_phase_retrieval(self, check):
        self._need_phase_retrieval = check


def is_valid_for_reconstruction(
    scan: TomoScanBase, need_phase_retrieval: bool = True, check_values: bool = False
):
    """
    check `scan` contains necessary and valid information to be reconstructed.

    :param TomoScanBase scan: scan to be checked
    :param check_values: If true check data for phase retrieval (energy, sample/detector distance...)
    :param check_datasets: open datasets to check for nan values or broken links to file
    """
    checker = ReconstructionValidator(
        scan=scan,
        check_phase_retrieval=need_phase_retrieval,
        check_values=check_values,
    )
    return checker.is_valid()



"""test of the .vol Volume"""

import os

import numpy
import pytest
from silx.io.url import DataUrl

from tomoscan.esrf.identifier import RawVolumeIdentifier
from tomoscan.esrf.volume.mock import create_volume
from tomoscan.esrf.volume.rawvolume import RawVolume
from tomoscan.identifier import VolumeIdentifier

_data = create_volume(frame_dims=(20, 20), z_size=60)
_data = _data.astype(numpy.uint8)

metadata_info = {
    "NUM_X": 20,
    "NUM_Y": 20,
    "NUM_Z": 60,
    "voxelSize": 2.3e-2,
    "BYTEORDER": "LOWBYTEFIRST",
    "ValMin": _data.min(),
    "ValMax": _data.max(),
    "s1": -1.13,
    "s2": 8.19,
    "S1": -4.13,
    "S2": 5.15,
}

metadata_xml = {
    "idAc": "N_A_",
    "listSubVolume": {
        "SUBVOLUME_NAME": "toto_volume",
        "SIZEX": metadata_info["NUM_X"],
        "SIZEY": metadata_info["NUM_Y"],
        "SIZEZ": metadata_info["NUM_Z"],
        "ORIGINX": 1,
        "ORIGINY": 1,
        "ORIGINZ": 1,
        "DIM_REC": 14155776000,
        "voxelsize": metadata_info["voxelSize"],
        "BYTE_ORDER": metadata_info["BYTEORDER"],
        "ValMin": metadata_info["ValMin"],
        "ValMax": metadata_info["ValMax"],
        "s1": metadata_info["s1"],
        "s2": metadata_info["s2"],
        "S1": metadata_info["S1"],
        "S2": metadata_info["S2"],
    },
}


def test_raw_volume(tmp_path):
    test_dir = str(tmp_path / "test_volume_url")
    os.makedirs(test_dir)
    volume_file = os.path.join(test_dir, "volume.vol")
    metadata_info_volume_file = os.path.join(test_dir, "volume.vol.info")
    metadata_xml_volume_file = os.path.join(test_dir, "volume.vol.xml")

    volume = RawVolume(
        file_path=volume_file,
        data=_data,
        metadata=metadata_info,
    )
    assert volume.url.file_path() == volume_file

    assert volume.url.data_path() is None
    assert volume.data_url.scheme() == "raw"
    assert volume.data_url.file_path() == volume_file
    assert volume.data_url.data_path() is None
    assert volume.metadata_url.scheme() == "info"
    assert volume.metadata_url.file_path() == metadata_info_volume_file
    assert volume.metadata_url.data_path() is None

    volume_identifier = volume.get_identifier()
    assert isinstance(volume_identifier.short_description(), str)
    assert isinstance(volume_identifier, VolumeIdentifier)

    assert not os.path.exists(volume_file)
    assert not os.path.exists(metadata_info_volume_file)
    assert not os.path.exists(metadata_xml_volume_file)

    # make sure saving other file format than float32 make it fails
    with pytest.raises(TypeError):
        volume.save()
    data_cast = volume.data.astype(numpy.float32)
    volume.data = data_cast
    volume.save()
    # try to save it twice to make sure we are not appending data to it

    assert os.path.exists(volume_file)
    assert os.path.exists(metadata_info_volume_file)
    assert not os.path.exists(metadata_xml_volume_file)
    xml_url = DataUrl(
        file_path=metadata_xml_volume_file,
        scheme="lxml",
    )
    volume.metadata = metadata_xml
    volume.save_metadata(url=xml_url)
    assert os.path.exists(metadata_xml_volume_file)

    assert tuple(volume.browse_data_urls()) == (volume.data_url,)
    assert tuple(volume.browse_data_files()) == (volume.data_url.file_path(),)
    assert tuple(volume.browse_metadata_files()) == (volume.metadata_url.file_path(),)

    volume_loaded = RawVolume(file_path=volume_file)
    assert volume.url.file_path() == volume_file
    assert volume.metadata_url.file_path() == metadata_info_volume_file
    data_no_cache = volume_loaded.load_data(volume_loaded.data_url, store=False)
    metadata_no_cache = volume_loaded.load_metadata(
        volume_loaded.metadata_url, store=False
    )
    assert data_no_cache is not None
    assert metadata_no_cache is not None
    assert volume_loaded.data is None
    assert volume_loaded.metadata is None
    data_float_32 = _data.astype(numpy.float32)
    numpy.testing.assert_array_almost_equal(data_float_32, data_no_cache)
    # only check keys because values are not cast (once read everything is a str...)
    assert metadata_info.keys() == metadata_no_cache.keys()
    assert volume.get_min_max_values() == (data_float_32.min(), data_float_32.max())
    # check error is raised if byte order is incorrect
    highbytefirst = numpy.dtype(">" + volume.data.dtype.char)
    volume.data = volume.data.astype(highbytefirst)
    with pytest.raises(TypeError):
        volume.save_data()

    # get the volume from the identifier
    volume_from_id = RawVolume.from_identifier(identifier=volume_identifier)
    assert volume_from_id.data is None
    assert volume_from_id.metadata is None

    # get the volume from the identifier str
    volume_from_str = RawVolume.from_identifier(
        identifier=RawVolumeIdentifier.from_str(volume_identifier.to_str())
    )
    assert volume_from_str.data is None
    assert volume_from_str.metadata is None
    assert volume_from_str.get_identifier() == volume_from_id.get_identifier()

    # test browsing frames
    # TODO

    # test hash
    hash(volume_from_str)
    assert volume.data_extension == "vol"
    assert volume.metadata_extension == "info"


def test_data_file_saver_generator(tmp_path):
    """
    data_file_saver_generator (dumping frame by frame a volume to disk)
    """
    volume_file_path = str(tmp_path / "volume_for_file_generator.vol")
    volume = RawVolume(file_path=volume_file_path)
    data_cast = _data.astype(numpy.float32)

    for slice_, slice_saver in zip(
        data_cast,
        volume.data_file_saver_generator(
            n_frames=_data.shape[0], data_url=volume.data_url, overwrite=False
        ),
    ):
        slice_saver[:] = slice_
    assert volume.data is None
    volume.metadata = metadata_info
    volume.save_metadata()
    numpy.testing.assert_array_equal(
        volume.load_data(),
        data_cast,
    )


def test_raw_volume_slice_getter(tmp_path):
    """
    tset slice getters
    """
    volume_file = str(tmp_path / "volume_test_getter.vol")
    data_cast = _data.astype(numpy.float32)

    volume = RawVolume(
        file_path=volume_file,
        data=data_cast,
        metadata=metadata_info,
        overwrite=True,
    )
    volume.save()
    assert os.path.exists(volume_file)

    volume_loaded = RawVolume(
        file_path=volume_file,
    )
    volume_loaded.load()
    numpy.testing.assert_array_equal(
        data_cast,
        volume_loaded.data,
    )

    # test get_slice function
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=0),
        data_cast[2],
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=1),
        data_cast[:, 2, :],
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=2),
        data_cast[:, :, 2],
    )
    volume_loaded.data = None

    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=0),
        data_cast[2],
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=1),
        data_cast[:, 2, :],
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=2),
        data_cast[:, :, 2],
    )


def test_append_writes(tmp_path):
    """
    Test trying to mimic PyHST2 output files that can be opened with imagej:
      - One single .vol, written in several steps (append mode)
      - One .info file (for users ?)
      - One .xml file (for ImageJ)
    """

    def generate_metadata(data, **kwargs):
        n_z, n_y, n_x = data.shape
        metadata = {
            "NUM_X": n_x,
            "NUM_Y": n_y,
            "NUM_Z": n_z,
            "voxelSize": kwargs.get("voxelSize", 40.0),
            "BYTEORDER": "LOWBYTEFIRST",
            "ValMin": kwargs.get("ValMin", 0.0),
            "ValMax": kwargs.get("ValMin", 1.0),
            "s1": 0.0,
            "s2": 0.0,
            "S1": 0.0,
            "S2": 0.0,
        }
        return metadata

    def sanitize_metadata(metadata):
        # To be fixed in RawVolume
        for what in ["NUM_X", "NUM_Y", "NUM_Z"]:
            metadata[what] = int(metadata[what])
        for what in ["voxelSize", "ValMin", "ValMax", "s1", "s2", "S1", "S2"]:
            metadata[what] = float(metadata[what])

    def write_data(volume, data):
        existing_metadata = volume.load_metadata()
        new_metadata = generate_metadata(data)
        if len(existing_metadata) == 0:
            # first write
            metadata = new_metadata
        else:
            # append write ; update metadata
            metadata = existing_metadata.copy()
            sanitize_metadata(metadata)
            metadata["NUM_Z"] += new_metadata["NUM_Z"]
        volume.data = data
        volume.metadata = metadata
        volume.save()
        # Also save .xml
        volume.save_metadata(
            url=DataUrl(
                scheme="lxml",
                file_path=volume.metadata_url.file_path().replace(".info", ".xml"),
            )
        )

    volume_file = str(tmp_path / "volume_append.vol")

    volume = RawVolume(
        file_path=volume_file,
        append=True,
    )

    # First write
    data_1 = numpy.arange(10 * 32 * 32, dtype="f").reshape((10, 32, 32))
    write_data(volume, data_1)

    # Second write - metadata should be updated
    data_2 = numpy.arange(11 * 32 * 32, dtype="f").reshape((11, 32, 32))
    write_data(volume, data_2)

    # Third write - metadata should be updated
    data_3 = numpy.arange(12 * 32 * 32, dtype="f").reshape((12, 32, 32))
    write_data(volume, data_3)

    assert int(volume.load_metadata()["NUM_Z"]) == 10 + 11 + 12
    assert os.path.getsize(volume_file) == data_1.nbytes + data_2.nbytes + data_3.nbytes


def test_example():
    """test static function 'example'"""
    assert isinstance(RawVolume.example_defined_from_str_identifier(), str)



# coding: utf-8

from __future__ import annotations

import logging
import os
import re
from glob import glob

import fabio
import numpy
from lxml import etree

from tomoscan.framereducer.target import REDUCER_TARGET
from tomoscan.framereducer.framereducerbase import FrameReducerBase
from tomoscan.framereducer.method import ReduceMethod
from tomoscan.scanbase import ReducedFramesInfos, TomoScanBase
from tomoscan.esrf.scan.utils import get_parameters_frm_par_or_info

_logger = logging.getLogger(__name__)


__all__ = [
    "EDFFrameReducer",
]


class EDFFrameReducer(FrameReducerBase):
    RAW_FLAT_RE = "ref*.*[0-9]{3,4}_[0-9]{3,4}"
    """regular expression to discover flat files"""

    RAW_DARK_RE = "darkend[0-9]{3,4}"
    """regular expression to discover raw dark files"""

    REFHST_PREFIX = "refHST"

    DARKHST_PREFIX = "dark.edf"

    def __init__(
        self,
        scan: TomoScanBase,
        reduced_method: ReduceMethod,
        target: REDUCER_TARGET,
        output_dtype: numpy.dtype | None,
        input_flat_pattern=RAW_FLAT_RE,
        input_dark_pattern=RAW_DARK_RE,
        flat_output_prefix=REFHST_PREFIX,
        dark_output_prefix=DARKHST_PREFIX,
        overwrite=False,
        file_ext=".edf",
    ):
        super().__init__(
            scan, reduced_method, target, overwrite=overwrite, output_dtype=output_dtype
        )
        self._input_flat_pattern = input_flat_pattern
        self._input_dark_pattern = input_dark_pattern
        self._dark_output_prefix = dark_output_prefix
        self._flat_output_prefix = flat_output_prefix
        self._file_ext = file_ext

    @property
    def input_flat_pattern(self):
        return self._input_flat_pattern

    @property
    def input_dark_pattern(self):
        return self._input_dark_pattern

    @staticmethod
    def _getInformation(scan, refFile, information, _type, aliases=None):
        """
        Parse files contained in the given directory to get the requested
        information

        :param scan: directory containing the acquisition. Must be an absolute path
        :param refFile: the refXXXX_YYYY which should contain information about the
                        scan.
        :return: the requested information or None if not found
        """

        def parseRefFile(filePath):
            header = fabio.open(filePath).header
            for k in aliases:
                if k in header:
                    return _type(header[k])
            return None

        def parseXMLFile(filePath):
            try:
                for alias in info_aliases:
                    tree = etree.parse(filePath)
                    elmt = tree.find("acquisition/" + alias)
                    if elmt is None:
                        continue
                    else:
                        info = _type(elmt.text)
                        if info == -1:
                            return None
                        else:
                            return info
            except etree.XMLSyntaxError as e:
                _logger.warning(e)
                return None

        def parseInfoFile(filePath):
            metadata = get_parameters_frm_par_or_info(filePath)
            # convert metadata to lower case
            for alias in info_aliases:
                if alias in metadata:
                    return _type(metadata[alias])
            return None

        info_aliases = [information.lower()]
        if aliases is not None:
            assert type(aliases) in (tuple, list)
            [info_aliases.append(alias.lower()) for alias in aliases]

        if not os.path.isdir(scan):
            return None

        if refFile is not None and os.path.isfile(refFile):
            try:
                info = parseRefFile(refFile)
            except IOError as e:
                _logger.warning(e)
            else:
                if info is not None:
                    return info

        baseName = os.path.basename(scan)
        infoFiles = [os.path.join(scan, baseName + ".info")]
        infoOnDataVisitor = infoFiles[0].replace("lbsram", "", 1)
        # hack to check in lbsram, would need to be removed to add some consistency
        if os.path.isfile(infoOnDataVisitor):
            infoFiles.append(infoOnDataVisitor)
        for infoFile in infoFiles:
            if os.path.isfile(infoFile) is True:
                info = parseInfoFile(infoFile)
                if info is not None:
                    return info

        xmlFiles = [os.path.join(scan, baseName + ".xml")]
        xmlOnDataVisitor = xmlFiles[0].replace("lbsram", "", 1)
        # hack to check in lbsram, would need to be removed to add some consistency
        if os.path.isfile(xmlOnDataVisitor):
            xmlFiles.append(xmlOnDataVisitor)
        for xmlFile in xmlFiles:
            if os.path.isfile(xmlFile) is True:
                info = parseXMLFile(xmlFile)
                if info is not None:
                    return info

        return None

    @staticmethod
    def getDARK_N(scan):
        return EDFFrameReducer._getInformation(
            os.path.abspath(scan),
            refFile=None,
            information="DARK_N",
            _type=int,
            aliases=["dark_N"],
        )

    @staticmethod
    def getTomo_N(scan):
        return EDFFrameReducer._getInformation(
            os.path.abspath(scan),
            refFile=None,
            information="TOMO_N",
            _type=int,
            aliases=["tomo_N"],
        )

    @staticmethod
    def get_closest_SR_current(scan_dir, refFile=None):
        """
        Parse files contained in the given directory to get information about the
        incoming energy for the serie `iSerie`

        :param scan_dir: directory containing the acquisition
        :param refFile: the refXXXX_YYYY which should contain information about the
                        energy.
        :return: the energy in keV or none if no energy found
        """
        return EDFFrameReducer._getInformation(
            os.path.abspath(scan_dir),
            refFile,
            information="SrCurrent",
            aliases=["SRCUR", "machineCurrentStart"],
            _type=float,
        )

    @staticmethod
    def get_closest_count_time(scan_dir, refFile=None):
        return EDFFrameReducer._getInformation(
            os.path.abspath(scan_dir),
            refFile,
            information="Count_time",
            aliases=tuple(),
            _type=float,
        )

    def get_info(self, keyword: str):
        with open(self.infofile) as file:
            infod = file.readlines()
            for line in infod:
                if keyword in line:
                    return int(line.split("=")[1])
        # not found:
        return 0

    def run(self) -> dict:
        self._raw_darks = []
        self._raw_flats = []
        infos = ReducedFramesInfos()
        directory = self.scan.path
        res = {}
        if not self.preprocess():
            _logger.warning(f"preprocessing of {self.scan} failed")
        else:
            _logger.info(f"start proccess darks and flat fields for {self.scan}")
            if self.reduced_method is ReduceMethod.NONE:
                return None
            shape = fabio.open(self.filelist_fullname[0]).shape
            for i in range(len(self.serievec)):
                largeMat = numpy.zeros(
                    (self.nframes * self.nFilePerSerie, shape[0], shape[1])
                )

                if (
                    self.reducer_target is REDUCER_TARGET.DARKS
                    and len(self.serievec) == 1
                ):
                    fileName = self.out_prefix
                    if fileName.endswith(self._file_ext) is False:
                        fileName = fileName + self._file_ext
                else:
                    fileName = (
                        self.out_prefix.rstrip(self._file_ext)
                        + self.serievec[i]
                        + self._file_ext
                    )
                fileName = os.path.join(directory, fileName)
                if os.path.isfile(fileName):
                    if self.overwrite is False:
                        _logger.info(f"skip creation of {fileName}, already existing")
                        continue

                if self.nFilePerSerie == 1:
                    fSerieName = os.path.join(directory, self.series[i])
                    header = {"method": f"{self.reduced_method.name} on 1 image"}
                    header["SRCUR"] = self.get_closest_SR_current(
                        scan_dir=directory, refFile=fSerieName
                    )
                    header["Count_time"] = self.get_closest_count_time(
                        scan_dir=directory,
                        refFile=fSerieName,
                    )

                    if self.nframes == 1:
                        largeMat[0] = fabio.open(fSerieName).data
                    else:
                        handler = fabio.open(fSerieName)
                        dShape = (self.nframes, handler.dim2, handler.dim1)
                        largeMat = numpy.zeros(dShape)
                        for iFrame in range(self.nframes):
                            largeMat[iFrame] = handler.getframe(iFrame).data
                else:
                    header = {
                        "method": self.reduced_method.name
                        + " on %d images" % self.nFilePerSerie
                    }
                    header["SRCUR"] = self.get_closest_SR_current(
                        scan_dir=directory, refFile=self.series[i][0]
                    )
                    header["Count_time"] = self.get_closest_count_time(
                        scan_dir=directory,
                        refFile=self.series[i][0],
                    )
                    for j, fName in zip(
                        range(self.nFilePerSerie), self.filesPerSerie[self.serievec[i]]
                    ):
                        file_BigMat = fabio.open(fName)
                        if self.nframes > 1:
                            for fr in range(self.nframes):
                                jfr = fr + j * self.nframes
                                largeMat[jfr] = file_BigMat.getframe(fr).getData()
                        else:
                            largeMat[j] = file_BigMat.data

                # update electrical machine current
                if header["SRCUR"] is not None:
                    if infos.machine_current is None:
                        infos.machine_current = []
                    infos.machine_current.append(header["SRCUR"])
                if header["Count_time"] is not None:
                    if infos.count_time is None:
                        infos.count_time = []
                    infos.count_time.append(header["Count_time"])

                if self.reduced_method is ReduceMethod.MEDIAN:
                    data = numpy.median(largeMat, axis=0)
                elif self.reduced_method is ReduceMethod.MEAN:
                    data = numpy.mean(largeMat, axis=0)
                elif self.reduced_method is ReduceMethod.FIRST:
                    data = largeMat[0]
                elif self.reduced_method is ReduceMethod.LAST:
                    data = largeMat[-1]
                elif self.reduced_method is ReduceMethod.NONE:
                    return
                else:
                    raise ValueError(
                        f"Unrecognized calculation type request {self.reduced_method}"
                    )

                if (
                    self.reducer_target is REDUCER_TARGET.DARKS and self.nacq > 1
                ):  # and self.nframes == 1:
                    nacq = self.getDARK_N(directory) or 1
                    data = data / nacq
                if self.output_dtype is not None:
                    data = data.astype(self.output_dtype)
                file_desc = fabio.edfimage.EdfImage(data=data, header=header)
                res[int(self.serievec[i])] = data
                i += 1
                file_desc.write(fileName)
            _logger.info("end proccess darks and flat fields")
        return res, infos

    def preprocess(self):
        # start setup function
        if self.reduced_method is ReduceMethod.NONE:
            return False
        if self.reducer_target is REDUCER_TARGET.DARKS:
            self.out_prefix = self._dark_output_prefix
            self.info_nacq = "DARK_N"
        else:
            self.out_prefix = self._flat_output_prefix
            self.info_nacq = "REF_N"

        # init
        self.nacq = 0
        """Number of acquisition runned"""
        self.files = 0
        """Ref or dark files"""
        self.nframes = 1
        """Number of frame per ref/dark file"""
        self.serievec = ["0000"]
        """List of series discover"""
        self.filesPerSerie = {}
        """Dict with key the serie id and values list of files to compute
        for median or mean"""
        self.infofile = ""
        """info file of the acquisition"""

        # sample/prefix and info file
        directory = self.scan.path
        self.prefix = os.path.basename(directory)
        extensionToTry = (".info", "0000.info")
        for extension in extensionToTry:
            infoFile = os.path.join(directory, self.prefix + extension)
            if os.path.exists(infoFile):
                self.infofile = infoFile
                break

        if self.infofile == "":
            _logger.debug(f"fail to found .info file for {self.scan}")

        """
        Set filelist
        """
        # do the job only if not already done and overwrite not asked
        self.out_files = sorted(glob(directory + os.sep + "*." + self._file_ext))

        self.filelist_fullname = self.get_originals()
        self.fileNameList = []
        [
            self.fileNameList.append(os.path.basename(_file))
            for _file in self.filelist_fullname
        ]
        self.fileNameList = sorted(self.fileNameList)
        self.nfiles = len(self.filelist_fullname)
        # if nothing to process
        if self.nfiles == 0:
            _logger.info(
                f"no {self.reducer_target} for {directory}, because no file to compute found"
            )
            return False

        self.fid = fabio.open(self.filelist_fullname[0])
        self.nframes = self.fid.nframes
        self.nacq = 0
        # get the info of number of acquisitions
        if self.infofile != "":
            self.nacq = self.get_info(self.info_nacq)

        if self.nacq == 0:
            self.nacq = self.nfiles

        self.nseries = 1
        if self.nacq > self.nfiles:
            # get ready for accumulation and/or file multiimage?
            self.nseries = self.nfiles
        if (
            self.nacq < self.nfiles
            and self.get_n_digits(self.fileNameList[0], directory=directory) < 2
        ):
            self.nFilePerSerie = self.nseries
            self.serievec, self.filesPerSerie = self.preprocess_PCOTomo()
        else:
            self.series = self.fileNameList
            self.serievec = self.get_series_value(self.fileNameList, self._file_ext)
            self.filesPerSerie, self.nFilePerSerie = self.group_files_per_serie(
                self.filelist_fullname, self.serievec
            )

        if self.filesPerSerie is not None:
            for serie in self.filesPerSerie:
                for _file in self.filesPerSerie[serie]:
                    if self.reducer_target is REDUCER_TARGET.DARKS:
                        self._raw_darks.append(os.path.join(self.scan.path, _file))
                    if self.reducer_target is REDUCER_TARGET.FLATS:
                        self._raw_flats.append(os.path.join(self.scan.path, _file))

        return self.serievec is not None and self.filesPerSerie is not None

    @staticmethod
    def get_series_value(fileNames, file_ext):
        assert len(fileNames) > 0
        is_there_digits = len(re.findall(r"\d+", fileNames[0])) > 0
        series = set()
        i = 0
        for fileName in fileNames:
            if is_there_digits:
                name = fileName.rstrip(file_ext)
                file_index = name.split("_")[-1]
                rm_not_numeric = re.compile(r"[^\d.]+")
                file_index = rm_not_numeric.sub("", file_index)
                series.add(file_index)
            else:
                series.add("%04d" % i)
                i += 1
        return list(series)

    @staticmethod
    def group_files_per_serie(files, series):
        def findFileEndingWithSerie(poolFiles, serie):
            res = []
            for _file in poolFiles:
                _f = _file.rstrip(".edf")
                if _f.endswith(serie):
                    res.append(_file)
            return res

        def checkSeriesFilesLength(serieFiles):
            length = -1
            for serie in serieFiles:
                if length == -1:
                    length = len(serieFiles[serie])
                elif len(serieFiles[serie]) != length:
                    _logger.error("Series with inconsistant number of ref files")

        assert len(series) > 0
        if len(series) == 1:
            return {series[0]: files}, len(files)
        assert len(files) > 0

        serieFiles = {}
        unattributedFiles = files.copy()
        for serie in series:
            serieFiles[serie] = findFileEndingWithSerie(unattributedFiles, serie)
            [unattributedFiles.remove(_f) for _f in serieFiles[serie]]

        if len(unattributedFiles) > 0:
            _logger.error(f"Failed to associate {unattributedFiles} to any serie")
            return {}, 0

        checkSeriesFilesLength(serieFiles)

        return serieFiles, len(serieFiles[list(serieFiles.keys())[0]])

    @staticmethod
    def get_n_digits(_file, directory):
        file_without_scanID = _file.replace(os.path.basename(directory), "", 1)
        return len(re.findall(r"\d+", file_without_scanID))

    def preprocess_PCOTomo(self):
        filesPerSerie = {}
        if self.nfiles % self.nacq == 0:
            assert self.nacq < self.nfiles
            self.nseries = self.nfiles // self.nacq
            self.series = self.fileNameList
        else:
            _logger.warning("Fail to deduce series")
            return None, None

        linear = (
            self.get_n_digits(self.fileNameList[0], directory=self.scan.scan_path) < 2
        )
        if linear is False:
            # which digit pattern contains the file number?
            lastone = True
            penulti = True
            for first_files in range(self.nseries - 1):
                digivec_1 = re.findall(r"\d+", self.fileNameList[first_files])
                digivec_2 = re.findall(r"\d+", self.fileNameList[first_files + 1])
                if lastone:
                    lastone = (int(digivec_2[-1]) - int(digivec_1[-1])) == 0
                if penulti:
                    penulti = (int(digivec_2[-2]) - int(digivec_1[-2])) == 0

            linear = not penulti

        if linear is False:
            digivec_1 = re.findall(r"\d+", self.fileNameList[self.nseries - 1])
            digivec_2 = re.findall(r"\d+", self.fileNameList[self.nseries])
            # confirm there is 1 increment after self.nseries in the uperlast last digit patern
            if (int(digivec_2[-2]) - int(digivec_1[-2])) != 1:
                linear = True

        # series are simple sublists in main filelist
        # self.series = []
        if linear is True:
            is_there_digits = len(re.findall(r"\d+", self.fileNameList[0])) > 0
            if is_there_digits:
                serievec = set([re.findall(r"\d+", self.fileNameList[0])[-1]])
            else:
                serievec = set(["0000"])
            for i in range(self.nseries):
                if is_there_digits:
                    serie = re.findall(r"\d+", self.fileNameList[i * self.nacq])[-1]
                    serievec.add(serie)
                    filesPerSerie[serie] = self.fileNameList[
                        i * self.nacq : (i + 1) * self.nacq
                    ]
                else:
                    serievec.add("%04d" % i)
            # in the sorted filelist, the serie is incremented, then the acquisition number:
        else:
            self.series = self.fileNameList[0 :: self.nseries]
            serievec = set([re.findall(r"\d+", self.fileNameList[0])[-1]])
            for serie in serievec:
                filesPerSerie[serie] = self.fileNameList[0 :: self.nseries]
        serievec = list(sorted(serievec))

        if len(serievec) > 2:
            _logger.error(
                f"DarkRefs do not deal with multiple scan. (scan {self.scan})"
            )
            return None, None
        assert len(serievec) <= 2
        if len(serievec) > 1:
            key = serievec[-1]
            tomoN = self.getTomo_N(self.scan)
            if tomoN is None:
                _logger.error("Fail to found information TOMO_N")
            del serievec[-1]
            serievec.append(str(tomoN).zfill(4))
            filesPerSerie[serievec[-1]] = filesPerSerie[key]
            del filesPerSerie[key]
            assert len(serievec) == 2
            assert len(filesPerSerie) == 2

        return serievec, filesPerSerie

    def get_originals(self) -> list:
        """compute the list of originals files to be used to compute the reducer target."""
        if self.reducer_target is REDUCER_TARGET.FLATS:
            try:
                pattern = re.compile(self.input_flat_pattern)
            except Exception:
                pattern = None
                _logger.error(
                    f"Fail to compute regular expresion for {self.input_flat_pattern}"
                )
        elif self.reducer_target is REDUCER_TARGET.DARKS:
            re.compile(self.input_dark_pattern)
            try:
                pattern = re.compile(self.input_dark_pattern)
            except Exception:
                pattern = None
                _logger.error(
                    f"Fail to compute regular expresion for {self.input_dark_pattern}"
                )
        else:
            raise ValueError(f"reducer target not handled ({self.reducer_target})")

        filelist_fullname = []
        if pattern is None:
            return filelist_fullname
        directory = self.scan.path
        for file in os.listdir(directory):
            if pattern.match(file) and file.endswith(self._file_ext):
                if (
                    file.startswith(self._flat_output_prefix)
                    or file.startswith(self._dark_output_prefix)
                ) is False:
                    filelist_fullname.append(os.path.join(directory, file))
        return sorted(filelist_fullname)

    def remove_raw_files(self):
        """Remove orignals files fitting the target (dark or flat files)"""
        if self.reducer_target is REDUCER_TARGET.DARKS:
            # In the case originals has already been found for the median
            # calculation
            if len(self._raw_darks) > 0:
                files = self._raw_darks
            else:
                files = self.get_originals()
        elif self.reducer_target is REDUCER_TARGET.FLATS:
            if len(self._raw_flats) > 0:
                files = self._raw_flats
            else:
                files = self.get_originals()
        else:
            _logger.error(
                f"the requested what (reduce {self.reducer_target}) is not recognized. "
                "Can't remove corresponding file"
            )
            return

        _files = set(files)
        for _file in _files:
            try:
                os.remove(_file)
            except Exception as e:
                _logger.error(e)



"""
fscan: parsing Bliss "fscans": https://gitlab.esrf.fr/bliss/fscan
"""

from __future__ import annotations

import posixpath
from os import path
from datetime import datetime
from silx.io.url import DataUrl
from silx.io.utils import open as hdf5_open
from .h5utils import get_first_hdf5_entry, get_h5obj_value, get_hdf5_dataset_shape
from .fscan_scantypes import get_scan_metadata
from tomoscan.utils.hdf5 import get_data_sources
import logging

_logger = logging.getLogger(__file__)


def is_dataset_entry(entry):
    """
    filter all entries that doesn't contain any detector dataset. By 'design' all entries that doesn't ends with '.1'
    """
    return entry.endswith(".1")


def list_datasets(fname):
    """
    List the entries in the form X.1 in a file
    """
    with hdf5_open(fname) as f:
        entries = list(f.keys())
    entries = [entry for entry in entries if is_dataset_entry(entry)]

    # Sort entries
    def get_entry_scan_num(entry_name):
        return int(entry_name.split(".")[0])

    entries.sort(key=get_entry_scan_num)
    return entries


def list_datasets_with_attributes(fname, attrs, default=None):
    """
    List the entries in the form X.1 in a file with requested attributes.
    If attributes key doesn't exists then take 'default' value
    """
    results = {}
    with hdf5_open(fname) as f:
        entries = list(f.keys())
        entries = [entry for entry in entries if is_dataset_entry(entry)]
        for entry in entries:
            results[entry] = {}
            for attr in attrs:
                if attr not in f[entry]:
                    res = default
                else:
                    res = f[entry][attr][()]
                results[entry][attr] = res
    return results


class FscanDataset:
    """
    A simple class for parsing ESRF-Fscan datasets
    https://gitlab.esrf.fr/bliss/fscan
    """

    _default_detector_name = None
    _instrument_name = "instrument"
    guess_detector_name_if_missing = True

    def __init__(
        self,
        fname: str,
        entry: str | None = None,
        detector_name: str | None = None,
        on_missing_metadata: str = "raise",
    ):
        """
        Build a Dataset object. Each object is tied to only one entry.

        :param fname: Path to the HDF5 file.
        :param entry: HDF5 entry. If not provided, the first entry is taken.
        :param detector_name: Detector name
        :param on_missing_metadata: behavior when let missing metadata. Valid values are 'print' or 'raise'
        """
        self.fname = fname
        self.entry = entry or get_first_hdf5_entry(fname)
        self.detector_name = detector_name or self._default_detector_name
        if self.detector_name is None and self.guess_detector_name_if_missing:
            self.detector_name = guess_detector_name(self.fname, self.entry)
        self._virtual_sources = None

        self._check_file()
        self._get_toplevel_fields()
        self._get_data_info()
        self._get_metadata(on_missing_metadata)

    def _check_file(self):
        self.data_path = posixpath.join(
            self.entry, self._instrument_name, self.detector_name, "data"
        )
        with hdf5_open(self.fname) as f:
            if self.entry not in f:
                raise ValueError("No entry '%s' in file %s" % (self.entry, self.fname))
            if (
                self._instrument_name not in f[self.entry]
                or "measurement" not in f[self.entry]
            ):
                raise ValueError(
                    "%s or measurement not found in %s/%s"
                    % (self._instrument_name, self.fname, self.entry)
                )
            if self.data_path not in f:
                raise ValueError(
                    "Cannot access data %s in file %s" % (self.data_path, self.fname)
                )

    def _get_toplevel_fields(self):
        with hdf5_open(self.fname) as f:
            current_entry = f[self.entry]
            for name in ["start_time", "end_time", "title"]:
                val = get_string(get_h5obj_value(current_entry, name))
                if name in ["start_time", "end_time"]:
                    val = get_datetime(val)
                setattr(self, name, val)

    def _get_data_info(self):
        with hdf5_open(self.fname) as f:
            self.data_shape = f[self.data_path].shape
        self.dataset_hdf5_url = DataUrl(
            file_path=self.fname, data_path=self.data_path, scheme="silx"
        )

    def _get_generic_key(self, name, h5_path, default=None):
        val = getattr(self, name, None)
        if val is not None:
            return val
        h5_group, h5_name = posixpath.split(h5_path)
        with hdf5_open(self.fname) as f:
            res = get_h5obj_value(f[h5_group], h5_name, default=default)
        setattr(self, name, res)
        return res

    @property
    def exposure_time(self):
        """
        Get the exposure time in seconds
        """
        expotime_path = posixpath.join(
            posixpath.dirname(self.data_path), "acq_parameters", "acq_expo_time"
        )
        return self._get_generic_key("_exposure_time", expotime_path)

    def get_virtual_sources(
        self, remove_nonexisting: bool = True, force_recompute: bool = False
    ):
        """
        Return a dict with the virtual sources of the current dataset.

        :param remove_nonexisting: Whether to check that each target file actually exists, and possibly remove the non-existing files
        :param force_recompute: if False and already get virtual sources in cache return them
        """
        # Do only one recursion - otherwise call get_data_sources()
        if self._virtual_sources is not None and not force_recompute:
            return self._virtual_sources

        sources = get_data_sources(
            fname=self.fname, data_path=self.data_path, recursive=False
        )

        if remove_nonexisting:
            to_discard = [f for f in sources.keys() if not (path.exists(f))]
            if to_discard != []:
                print(
                    "Warning: Scan %s: the following files were declared in master file, but not found: %s"
                    % (
                        path.basename(self.fname) + ":" + self.entry,
                        str([path.basename(f) for f in to_discard]),
                    )
                )
                for fname in to_discard:
                    sources.pop(fname)

        self._virtual_sources = sources
        return sources

    def get_stack_size(self, use_file_n: int = 0):
        """
        Get dataset stack size of one LIMA file (size can be different, if cancel for example)

        :param use_file_n: Which file to take to get stack size. Default is first file.
        """
        virtual_sources = self.get_virtual_sources()
        fnames = list(virtual_sources.keys())

        fname = fnames[use_file_n]
        h5path = virtual_sources[fname]

        shp = get_hdf5_dataset_shape(fname, h5path)
        return shp[0]

    def get_all_stacks_sizes(self):
        """
        Go through all LIMA files and retrieve dataset size (nb frame)
        """
        virtual_sources = self.get_virtual_sources()
        stacks_sizes = []
        for fname, h5path in virtual_sources.items():
            shp = get_hdf5_dataset_shape(fname, h5path)
            stacks_sizes.append(shp[0])
        return stacks_sizes

    def _get_metadata(self, on_missing_metadata):
        self.metadata = get_scan_metadata(
            self.fname, self.entry, self.detector_name, on_error=on_missing_metadata
        )

    def _tostr(self):
        return str(
            "%s(fname=%s, entry=%s)" % (self.__class__.__name__, self.fname, self.entry)
        )

    def __str__(self):
        return self._tostr()

    def __repr__(self):
        return self._tostr()


def guess_detector_name(fname, entry):
    with hdf5_open(fname) as f:
        meas = f[entry]["measurement"]
        for k in meas.keys():
            if hasattr(meas[k], "ndim") and meas[k].ndim == 3:
                return k


def get_string(str_or_bytes):
    if isinstance(str_or_bytes, bytes):
        return str_or_bytes.decode()
    return str_or_bytes


def format_time(timestamp_str):
    try:
        d = datetime.fromisoformat(timestamp_str)
    except ValueError:
        if timestamp_str.endswith("Z"):
            # https://docs.python.org/3/library/datetime.html#technical-detail
            # providing 'Z' is identical to '+00:00'
            timestamp_str = timestamp_str.replace("Z", "+00:00")
            d = datetime.fromisoformat(timestamp_str)
    return d.isoformat()  # long format


def get_datetime(date_str: str) -> str | None:
    """
    try to format the date. If fail return None
    """
    if date_str is None:
        return None
    try:
        date = format_time(date_str)
    except Exception as exc:
        _logger.warning(f"fail to convert {date_str}. Error is {exc}")
        date = None
    return date



# coding: utf-8
"""module defining utils for a jp2k volume"""

from __future__ import annotations

import logging
import os

import numpy
from packaging.version import parse as parse_version
from silx.io.url import DataUrl

from tomoscan.esrf.identifier.jp2kidentifier import JP2KVolumeIdentifier
from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring
from tomoscan.utils.volume import rescale_data

from .singleframebase import VolumeSingleFrameBase

try:
    import glymur  # noqa #F401 needed for later possible lazy loading
except ImportError:
    has_glymur = False
    has_minimal_openjpeg = False
    glymur_version = None
    openjpeg_version = None
else:
    has_glymur = True
    from glymur import set_option as glymur_set_option
    from glymur.version import openjpeg_version
    from glymur.version import version as glymur_version

    if openjpeg_version < "2.3.0":
        has_minimal_openjpeg = False
    else:
        has_minimal_openjpeg = True

_logger = logging.getLogger(__name__)

_MISSING_GLYMUR_MSG = "Fail to import glymur. won't be able to load / save volume to jp2k. You can install it by calling pip."

__all__ = [
    "JP2KVolume",
    "has_glymur",
    "has_minimal_openjpeg",
    "glymur_version",
    "openjpeg_version",
    "get_available_threads",
]


class JP2KVolume(VolumeSingleFrameBase):
    """
    Save volume data to single frame jp2k files and metadata to .txt file

    :param cratios: list of ints. compression ratio for each jpeg2000 layer
    :param psnr: list of int.
                                The PSNR (Peak Signal-to-Noise ratio) for each jpeg2000 layer.
                                This defines a quality metric for lossy compression.
                                The number "0" stands for lossless compression.
    :param n_threads: number of thread to use for writing. If None will try to get as much as possible
    :param clip_values: optional tuple of two float (min, max) to clamp volume value
    :param rescale_data: rescale data before dumping each frame. Expected to be True when dump a new volume and False
                              when save volume cast for example (and when histogram is know...)

    :warning: each file saved under {volume_basename}_{index_zfill6}.jp2k is considered to be a slice of the volume.
    """

    DEFAULT_DATA_EXTENSION = "jp2"

    DEFAULT_DATA_SCHEME = "glymur"

    def __init__(
        self,
        folder: str | None = None,
        volume_basename: str | None = None,
        data: numpy.ndarray | None = None,
        source_scan: TomoScanBase | None = None,
        metadata: dict | None = None,
        data_url: DataUrl | None = None,
        metadata_url: DataUrl | None = None,
        overwrite: bool = False,
        start_index=0,
        data_extension=DEFAULT_DATA_EXTENSION,
        metadata_extension=VolumeSingleFrameBase.DEFAULT_METADATA_EXTENSION,
        cratios: list | None = None,
        psnr: list | None = None,
        n_threads: int | None = None,
        clip_values: tuple | None = None,
        rescale_data: bool = True,
    ) -> None:
        if folder is not None:
            url = DataUrl(
                file_path=str(folder),
                data_path=None,
            )
        else:
            url = None
        super().__init__(
            url=url,
            data=data,
            volume_basename=volume_basename,
            source_scan=source_scan,
            metadata=metadata,
            data_url=data_url,
            metadata_url=metadata_url,
            overwrite=overwrite,
            start_index=start_index,
            data_extension=data_extension,
            metadata_extension=metadata_extension,
        )
        if not has_glymur:
            _logger.warning(_MISSING_GLYMUR_MSG)
        else:
            if not has_minimal_openjpeg:
                _logger.warning(
                    "You must have at least version 2.3.0 of OpenJPEG "
                    "in order to write jp2k images."
                )
        self._cratios = cratios
        self._psnr = psnr
        self._clip_values = None
        self._cast_already_logged = False
        """bool used to avoid logging potential data cast for each frame when save a volume"""
        self._rescale_data = rescale_data
        """should we rescale data before dumping the frame"""
        self.setup_multithread_encoding(n_threads=n_threads)
        self.clip_values = clip_values  # execute test about the type...

    @property
    def cratios(self) -> list | None:
        return self._cratios

    @cratios.setter
    def cratios(self, cratios: list | None):
        self._cratios = cratios

    @property
    def psnr(self) -> list | None:
        return self._psnr

    @psnr.setter
    def psnr(self, psnr: list | None):
        self._psnr = psnr

    @property
    def rescale_data(self) -> bool:
        return self._rescale_data

    @rescale_data.setter
    def rescale_data(self, rescale: bool) -> None:
        if not isinstance(rescale, bool):
            raise TypeError
        self._rescale_data = rescale

    @property
    def clip_values(self) -> tuple | None:
        """
        :return: optional min and max value to clip - as float.
        """
        return self._clip_values

    @clip_values.setter
    def clip_values(self, values: tuple | None) -> None:
        if values is None:
            self._clip_values = None
        elif not isinstance(values, (tuple, list)):
            raise TypeError
        elif not len(values) == 2:
            raise ValueError("clip values are expected to be two floats")
        elif not values[1] >= values[0]:
            raise ValueError
        else:
            self._clip_values = values

    @docstring(VolumeSingleFrameBase)
    def save_data(self, url: DataUrl | None = None) -> None:
        self._cast_already_logged = False
        super().save_data(url=url)

    @docstring(VolumeSingleFrameBase)
    def save_frame(self, frame, file_name, scheme):
        if not has_glymur:
            raise RuntimeError(_MISSING_GLYMUR_MSG)

        expected_output_dtype = frame.dtype
        if self.clip_values is not None:
            frame = numpy.clip(frame, self.clip_values[0], self.clip_values[1])
        if expected_output_dtype not in (numpy.uint8, numpy.uint16):
            _logger.info(
                f"{self.get_identifier().to_str()} get {frame.dtype}. Cast it as {numpy.uint16}"
            )
            expected_output_dtype = numpy.uint16
        if self.rescale_data:
            max_uint = numpy.iinfo(expected_output_dtype).max
            frame = rescale_data(
                data=frame,
                new_min=0,
                new_max=max_uint,
                data_min=self.clip_values[0] if self.clip_values is not None else None,
                data_max=self.clip_values[1] if self.clip_values is not None else None,
            )
        frame = frame.astype(expected_output_dtype)

        if scheme == "glymur":
            glymur.Jp2k(file_name, data=frame, psnr=self.psnr, cratios=self.cratios)
        else:
            raise ValueError(f"Scheme {scheme} is not handled")

    @docstring(VolumeSingleFrameBase)
    def load_frame(self, file_name, scheme):
        if not has_glymur:
            raise RuntimeError(_MISSING_GLYMUR_MSG)

        if scheme == "glymur":
            jp2_file = glymur.Jp2k(file_name)
            return jp2_file[:]
        else:
            raise ValueError(f"Scheme {scheme} is not handled")

    def read_file(self, file_name) -> tuple:
        jp2_file = glymur.Jp2k(file_name)
        return jp2_file[:]

    @staticmethod
    def setup_multithread_encoding(
        n_threads: int | None = None, what_if_not_available: str = "ignore"
    ):
        """
        Setup OpenJpeg multi-threaded encoding.

        :param n_threads: Number of threads. If not provided, all available threads are used.
        :param what_if_not_available: What to do if requirements are not fulfilled. Possible values are:
            - "ignore": do nothing, proceed
            - "print": show an information message
            - "raise": raise an error

        """
        required_glymur_version = "0.9.3"
        required_openjpeg_version = "2.4.0"

        def not_available(msg):
            if what_if_not_available == "raise":
                raise ValueError(msg)
            elif what_if_not_available == "print":
                print(msg)

        if not has_glymur:
            not_available(f"glymur not installed. {required_glymur_version} required")
            return
        elif parse_version(glymur_version) < parse_version(required_glymur_version):
            not_available(
                f"glymur >= {required_glymur_version} is required for multi-threaded encoding (current version: {glymur_version})"
            )
            return
        elif not has_minimal_openjpeg:
            not_available(
                f"libopenjpeg >= {required_openjpeg_version} is required for multi-threaded encoding (current version: {openjpeg_version})"
            )
            return

        if n_threads is None:
            n_threads = get_available_threads()
        glymur_set_option("lib.num_threads", n_threads)

    @staticmethod
    @docstring(VolumeSingleFrameBase)
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, JP2KVolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {JP2KVolumeIdentifier}"
            )
        return JP2KVolume(
            folder=identifier.folder,
            volume_basename=identifier.file_prefix,
        )

    @docstring(VolumeSingleFrameBase)
    def get_identifier(self) -> JP2KVolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")
        return JP2KVolumeIdentifier(
            object=self, folder=self.url.file_path(), file_prefix=self._volume_basename
        )

    @staticmethod
    def example_defined_from_str_identifier() -> str:
        return " ; ".join(
            [
                f"{JP2KVolume(folder='/path/to/my/my_folder').get_identifier().to_str()}",
                f"{JP2KVolume(folder='/path/to/my/my_folder', volume_basename='mybasename').get_identifier().to_str()} (if mybasename != folder name)",
            ]
        )


def get_available_threads():
    return len(os.sched_getaffinity(0))



# coding: utf-8
"""This modules contains base class for TomoScanBase"""

from __future__ import annotations

import pint
import logging
import os
import pathlib
from bisect import bisect_left
from collections import OrderedDict
from math import ceil
import warnings

import fabio
import h5py
import numpy
import silx.io.utils
from silx.io.url import DataUrl
from silx.io.utils import get_data
from silx.io.utils import open as hdf5_open
from tomoscan.utils.io import deprecated, deprecated_warning
from tomoscan.framereducer.ReduceFrameSaver import ReduceFrameSaver

from tomoscan.identifier import ScanIdentifier
from tomoscan.normalization import IntensityNormalization
from tomoscan.normalization import Method as _IntensityMethod
from tomoscan.normalization import normalize_chebyshev_2D, normalize_lsqr_spline_2D
from tomoscan.framereducer.reducedframesinfos import ReducedFramesInfos

from .progress import Progress
from .tomoobject import TomoObject
from .source import *  # noqa F401

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)

__all__ = [
    "TomoScanBase",
    "projections_angles_to_indices",
    "projection_angular_range_to_indices",
]


class TomoScanBase(TomoObject):
    """
    Base Class representing a scan.
    It is used to obtain core information regarding an acquisition like
    projections, dark and flat field...

    :param scan: path to the root folder containing the scan.
    """

    DICT_TYPE_KEY = "type"

    DICT_PATH_KEY = "path"

    _SCHEME = None
    """scheme to read data url for this type of acquisition"""

    FRAME_REDUCER_CLASS = None
    """Frame reducer class to be use in order to compute reduced darks and reduced flats"""

    def __init__(
        self,
        scan: str | None,
        type_: str,
        ignore_projections: dict | list | tuple | numpy.ndarray | None = None,
    ):
        super().__init__()
        self.path = scan
        self._type = type_
        self._reduced_flats = None
        """darks once reduced. We must have one per s. When set a dict is expected with index as the key
           and median or median of darks series as value"""
        self._reduced_flats_infos = ReducedFramesInfos()

        self._reduced_darks = None
        """flats once reduced.  We must have one per series. When set a dict is expected with index as the key
           and median or median of darks series as value"""
        self._reduced_darks_infos = ReducedFramesInfos()

        self._notify_ffc_rsc_missing = True
        """Should we notify the user if ffc fails because cannot find dark or
        flat. Used to avoid several warnings. Only display one"""
        self._projections = None
        self._alignment_projections = None
        self._flats_weights = None
        """list flats indexes to use for flat field correction and associate
        weights"""
        self._ignore_projections = ignore_projections
        self._configure_ignore_projections()
        """Extra information for normalization"""
        self._intensity_monitor = None
        """monitor of the intensity during acquisition. Can be a diode
        for example"""
        self._source = None
        self._intensity_normalization = IntensityNormalization()
        """Extra information for normalization"""
        self._machine_current = None
        self._count_time = None
        self._sample_detector_distance = None
        self._source_sample_distance = None
        self._propagation_distance: float | None = None
        self._source_type = None
        self._source_name = None
        self._instrument_name = None
        self._title = None

    def clear_cache(self):
        """clear caches. Might be call if some data changed after
        first read of data or metadata"""
        self._notify_ffc_rsc_missing = True
        self._source_type = None
        self._source_name = None
        self._instrument_name = None
        self._title = None
        self._sample_detector_distance = None
        self._source_sample_distance = None
        self._propagation_distance: float | None = None
        self.clear_frames_cache()

    @deprecated(replacement="clear_cache", since_version="2.2", reason="renaming")
    def clear_caches(self) -> None:
        self.clear_cache()

    def clear_frames_cache(self):
        self._alignment_projections = None
        self._flats_weights = None
        self._projections = None

    @deprecated(
        replacement="clear_frames_cache", since_version="2.2", reason="renaming"
    )
    def clear_frames_caches(self) -> None:
        self.clear_frames_cache()

    def _configure_ignore_projections(self):
        if self._ignore_projections is None:
            return
        usage_msg = (
            "Please provide a dictionary in the form {'kind': kind, 'values': values}"
        )
        if not (isinstance(self._ignore_projections, dict)):
            deprecated_warning(
                type_="Function",
                name="_configure_ignore_projections",
                reason=usage_msg,
                since_version="2.0.6",
            )
            self._ignore_projections = {
                "kind": "indices",
                "values": self._ignore_projections,
            }
        # Assuming dict from now on
        ignore_projections = self._ignore_projections
        if (
            "kind" not in ignore_projections
            or "values" not in ignore_projections
            or ignore_projections["kind"] not in ["indices", "angles", "range"]
        ):
            raise ValueError(usage_msg)
        if ignore_projections["kind"] == "indices" and numpy.array(
            ignore_projections["values"]
        ).dtype.kind not in ["u", "i"]:
            raise ValueError("Expected integer numbers for indices")

    @property
    def ignore_projections(self):
        return self._ignore_projections

    @ignore_projections.setter
    def ignore_projections(self, val):
        raise ValueError(
            "ignore_projections should not be modified after class initialization"
        )

    def get_ignored_projection_indices(self):
        if self.ignore_projections is None:
            return []
        ignore_kind = self.ignore_projections["kind"]
        if ignore_kind == "indices":
            return self.ignore_projections["values"]
        else:
            return self._get_excluded_projections_from_angles(ignore_kind)

    def _get_excluded_projections_from_angles(self, kind):
        if kind == "angles":
            mapping_func = projections_angles_to_indices
        elif kind == "range":
            mapping_func = projection_angular_range_to_indices
        else:
            raise ValueError
        projections_indices, rotation_angles = self._get_projs_indices_and_angles()
        return mapping_func(
            self.ignore_projections["values"], projections_indices, rotation_angles
        )

    def _get_projs_indices_and_angles(self):
        raise NotImplementedError("Base class")

    @property
    def reduced_darks(self):
        return self._reduced_darks

    def set_reduced_darks(
        self, darks, darks_infos: ReducedFramesInfos | dict | None = None
    ):
        self._reduced_darks = darks
        self.reduced_darks_infos = darks_infos

    @property
    def reduced_flats(self):
        return self._reduced_flats

    def set_reduced_flats(
        self, flats, flats_infos: ReducedFramesInfos | dict | None = None
    ):
        self._reduced_flats = flats
        self.reduced_flats_infos = flats_infos

    @property
    def reduced_darks_infos(self):
        return self._reduced_darks_infos

    @reduced_darks_infos.setter
    def reduced_darks_infos(self, infos: ReducedFramesInfos | dict | None):
        if infos is None:
            self._reduced_darks_infos.clear()
        elif isinstance(infos, ReducedFramesInfos):
            self._reduced_darks_infos = infos
        elif isinstance(infos, dict):
            self._reduced_darks_infos.load_from_dict(dict)
        else:
            raise TypeError

    @property
    def reduced_flats_infos(self):
        return self._reduced_flats_infos

    @reduced_flats_infos.setter
    def reduced_flats_infos(self, infos: ReducedFramesInfos | dict | None):
        if infos is None:
            self._reduced_flats_infos.clear()
        elif isinstance(infos, ReducedFramesInfos):
            self._reduced_flats_infos = infos
        elif isinstance(infos, dict):
            self._reduced_flats_infos.load_from_dict(dict)
        else:
            raise TypeError(f"unexpected error ({type(infos)})")

    @property
    def path(self) -> str | None:
        """

        :return: path of the scan root folder.
        """
        return self._path

    @path.setter
    def path(self, path: str | None) -> None:
        if path is None:
            self._path = path
        else:
            if not isinstance(path, (str, pathlib.Path)):
                raise TypeError(
                    f"path is expected to be a str or a pathlib.Path not {type(path)}"
                )
            self._path = os.path.abspath(str(path))

    @property
    def type(self) -> str:
        """

        :return: type of the scanBase (can be 'edf' or 'hdf5' for now).
        """
        return self._type

    @staticmethod
    def is_tomoscan_dir(directory: str, **kwargs) -> bool:
        """
        Check if the given directory is holding an acquisition

        :param directory:
        :return: does the given directory contains any acquisition
        """
        raise NotImplementedError("Base class")

    def is_abort(self, **kwargs) -> bool:
        """

        :return: True if the acquisition has been abort
        """
        raise NotImplementedError("Base class")

    @property
    def source(self):
        return self._source

    @property
    def flats(self) -> dict | None:
        """list of flats files"""
        return self._flats

    @flats.setter
    def flats(self, flats: dict | None) -> None:
        self._flats = flats

    @property
    def darks(self) -> dict | None:
        """list of darks files"""
        return self._darks

    @darks.setter
    def darks(self, darks: dict | None) -> None:
        self._darks = darks

    @property
    def projections(self) -> dict | None:
        """if found dict of projections urls with index during acquisition as
        key"""
        return self._projections

    @projections.setter
    def projections(self, projections: dict) -> None:
        self._projections = projections

    @property
    def alignment_projections(self) -> dict | None:
        """
        dict of projections made for alignment with acquisition index as key
        None if not found
        """
        return self._alignment_projections

    @alignment_projections.setter
    def alignment_projections(self, alignment_projs):
        self._alignment_projections = alignment_projs

    @property
    def dark_n(self) -> int | None:
        raise NotImplementedError("Base class")

    @property
    def tomo_n(self) -> int | None:
        """number of projection WITHOUT the return projections"""
        raise NotImplementedError("Base class")

    @property
    def flat_n(self) -> int | None:
        """number of flat per series (computed on the first series)"""
        raise NotImplementedError("Base class")

    @property
    def pixel_size(self) -> float | None:
        raise NotImplementedError("Base class")

    @property
    @deprecated(replacement="", since_version="1.1.0")
    def x_real_pixel_size(self) -> float | None:
        raise NotImplementedError("Base class")

    @property
    @deprecated(replacement="", since_version="1.1.0")
    def y_real_pixel_size(self) -> float | None:
        raise NotImplementedError("Base class")

    def get_pixel_size(self, unit="m") -> float | None:
        if self.pixel_size:
            return (self.pixel_size * _ureg.meter).to(unit).magnitude
        else:
            return None

    @property
    def instrument_name(self) -> str | None:
        """

        :return: instrument name
        """
        raise NotImplementedError("Base class")

    @property
    def dim_1(self) -> int | None:
        raise NotImplementedError("Base class")

    @property
    def dim_2(self) -> int | None:
        raise NotImplementedError("Base class")

    @property
    def ff_interval(self) -> int | None:
        raise NotImplementedError("Base class")

    @property
    def scan_range(self) -> int | None:
        raise NotImplementedError("Base class")

    @property
    def energy(self) -> float | None:
        """

        :return: incident beam energy in keV
        """
        raise NotImplementedError("Base class")

    @property
    def intensity_monitor(self):
        raise NotImplementedError("Base class")

    @property
    def sample_detector_distance(self) -> float | None:
        """

        :return: sample / detector distance in meter
        """
        raise NotImplementedError("Base class")

    @property
    @deprecated(replacement="sample_detector_distance", since_version="2.2.0")
    def distance(self) -> float | None:
        return self.sample_detector_distance

    @property
    def source_sample_distance(self) -> float | None:
        """

        :return: source / sample distance (in meter). Expected to be negative (NXtomo convention).
        """
        raise NotImplementedError("Base class")

    @property
    def field_of_view(self):
        """

        :return: field of view of the scan. None if unknown else Full or Half
        """
        raise NotImplementedError("Base class")

    @property
    def estimated_cor_frm_motor(self):
        deprecated_warning(
            type_="property",
            name="estimated_cor_frm_motor",
            replacement="x_rotation_axis_pixel_position",
            since_version="2.1",
        )
        return self.x_rotation_axis_pixel_position

    @property
    def x_rotation_axis_pixel_position(self):
        """

        :return: Estimated center of rotation estimated from motor position. In [-frame_width, +frame_width]. None if unable to find it
        """
        raise NotImplementedError("Base class")

    @property
    def x_translation(self) -> None | tuple:
        raise NotImplementedError("Base class")

    @property
    def y_translation(self) -> None | tuple:
        raise NotImplementedError("Base class")

    @property
    def z_translation(self) -> None | tuple:
        raise NotImplementedError("Base class")

    def get_sample_detector_distance(self, unit="m") -> float | None:
        """

        :param unit: unit requested for the distance
        :return: sample / detector distance with the requested unit
        """
        if self.sample_detector_distance:
            return (self.distance * _ureg.meter).to(unit).magnitude
        else:
            return None

    @deprecated(replacement="get_sample_detector_distance", since_version="2.0.0")
    def get_distance(self, *args, **kwargs):
        return self.get_sample_detector_distance(*args, **kwargs)

    @property
    @deprecated(
        replacement="sample_x_pixel_size", since_version="2.2", reason="renaming"
    )
    def x_pixel_size(self) -> float | None:
        return self.sample_x_pixel_size

    @property
    @deprecated(
        replacement="sample_y_pixel_size", since_version="2.2", reason="renaming"
    )
    def y_pixel_size(self) -> float | None:
        return self.sample_y_pixel_size

    @property
    def sample_x_pixel_size(self) -> float | None:
        raise NotImplementedError("Base class")

    @property
    def sample_y_pixel_size(self) -> float | None:
        raise NotImplementedError("Base class")

    @property
    def detector_x_pixel_size(self) -> float | None:
        raise NotImplementedError("Base class")

    @property
    def detector_y_pixel_size(self) -> float | None:
        raise NotImplementedError("Base class")

    @property
    def magnification(self) -> float | None:
        raise NotImplementedError("Base class")

    def update(self) -> None:
        """Parse the root folder and files to update information"""
        raise NotImplementedError("Base class")

    @property
    def sequence_name(self):
        """Return the sequence name"""
        raise NotImplementedError("Base class")

    @property
    def sample_name(self):
        """Return the sample name"""
        raise NotImplementedError("Base class")

    @property
    def propagation_distance(self) -> float | None:
        raise NotImplementedError("Base class")

    @property
    def group_size(self):
        """Used in the case of z-series for example. Return the number of
        sequence expected on the acquisition"""
        raise NotImplementedError("Base class")

    @property
    def count_time(self) -> list | None:
        raise NotImplementedError("Base class")

    @property
    def electric_current(self) -> tuple:
        warnings.warn(
            "electric_current is deprecated and will be removed in a future version. Use machine_current instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.machine_current

    @electric_current.setter
    def electric_current(self, current: tuple | None) -> None:
        warnings.warn(
            "electric_current is deprecated and will be removed in a future version. Use machine_current instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.machine_current = current

    @property
    def machine_current(self) -> tuple:
        """Return the sample name"""
        raise NotImplementedError("Base class")

    @machine_current.setter
    def machine_current(self, current: tuple | None) -> None:
        if not isinstance(current, (type(None), tuple)):
            raise TypeError(
                f"current is expected to be None or a tuple. Not {type(current)}"
            )
        self._machine_current = current

    @property
    def title(self):
        raise NotImplementedError("Base class")

    @property
    def source_name(self):
        raise NotImplementedError("Base class")

    @property
    def source_type(self):
        raise NotImplementedError("Base class")

    @property
    def x_flipped(self) -> bool:
        """
        warning: deprecated !!!!!
        return True if  the frames are flip through x
        """
        raise NotImplementedError("Base class")

    @property
    def y_flipped(self) -> bool:
        """
        warning: deprecated !!!!!
        return True if  the frames are flip through y
        """
        raise NotImplementedError("Base class")

    def get_x_flipped(self, default=None):
        deprecated_warning(
            type_="property",
            name="get_x_flipped",
            replacement="get_detector_transformations",
            since_version="1.3",
        )
        if self.x_flipped is None:
            return default
        else:
            return self.x_flipped

    def get_y_flipped(self, default=None):
        deprecated_warning(
            type_="property",
            name="get_y_flipped",
            replacement="get_detector_transformations",
            since_version="1.3",
        )
        if self.y_flipped is None:
            return default
        else:
            return self.y_flipped

    @property
    def detector_transformations(self) -> tuple | None:
        """
        return tuple of `Transformation` applied to the detector
        """
        raise NotImplementedError

    def get_detector_transformations(self, default):
        if self.detector_transformations is None:
            return default
        else:
            return self.detector_transformations

    def get_identifier(self) -> ScanIdentifier:
        """
        return the dataset identifier of the scan.
        The identifier is insure to be unique for each scan and
        allow the user to store the scan as a string identifier
        and to retrieve it later from this single identifier.
        """
        raise NotImplementedError("Base class")

    def to_dict(self) -> dict:
        """

        :return: convert the TomoScanBase object to a dictionary.
                 Used to serialize the object for example.
        """
        res = dict()
        res[self.DICT_TYPE_KEY] = self.type
        res[self.DICT_PATH_KEY] = self.path
        return res

    def load_from_dict(self, _dict: dict):
        """
        Load properties contained in the dictionary.

        :param _dict: dictionary to load
        :return: self
        :raises: ValueError if dict is invalid
        """
        raise NotImplementedError("Base class")

    def equal(self, other) -> bool:
        """

        :param instance to compare with
        :return: True if instance are equivalent

        ..note:: we cannot use the __eq__ function because this object need to be
                 picklable
        """
        return (
            isinstance(other, self.__class__)
            or isinstance(self, other.__class__)
            and self.type == other.type
            and self.path == other.path
        )

    def get_proj_angle_url(self) -> dict:
        """
        return a dictionary of all the projection. key is the angle of the
        projection and value is the url.

        Keys are int for 'standard' projections and strings for return
        projections.

        :return: angles as keys, radios as value.
        """
        raise NotImplementedError("Base class")

    @staticmethod
    def map_urls_on_scan_range(urls, n_projection, scan_range) -> dict:
        """
        map given urls to an angle regarding scan_range and number of projection.
        We take the hypothesis that 'extra projection' are taken regarding the
        'id19' policy:

         * If the acquisition has a scan range of 360 then:

            * if 4 extra projection, the angles are (270, 180, 90, 0)

            * if 5 extra projection, the angles are (360, 270, 180, 90, 0)

         * If the acquisition has a scan range of 180 then:

            * if 2 extra projections: the angles are (90, 0)

            * if 3 extra projections: the angles are (180, 90, 0)

        ..warning:: each url should contain only one radio.

        :param urls: dict with all the urls. First url should be
                     the first radio acquire, last url should match the last
                     radio acquire.
        :param n_projection: number of projection for the sample.
        :param scan_range: acquisition range (usually 180 or 360)
        :return: angle in degree as key and url as value

        :raises: ValueError if the number of extra images found and scan_range
                 are incoherent
        """
        assert n_projection is not None
        ordered_url = OrderedDict(sorted(urls.items(), key=lambda x: x))

        res = {}
        # deal with the 'standard' acquisitions
        for proj_i in range(n_projection):
            url = list(ordered_url.values())[proj_i]
            if n_projection == 1:
                angle = 0.0
            else:
                angle = proj_i * scan_range / (n_projection - 1)
            if proj_i < len(urls):
                res[angle] = url

        if len(urls) > n_projection:
            # deal with extra images (used to check if the sampled as moved for
            # example)
            extraImgs = list(ordered_url.keys())[n_projection:]
            if len(extraImgs) in (4, 5):
                if scan_range < 360:
                    _logger.warning(
                        "incoherent data information to retrieve"
                        "scan extra images angle"
                    )
                elif len(extraImgs) == 4:
                    res["270(1)"] = ordered_url[extraImgs[0]]
                    res["180(1)"] = ordered_url[extraImgs[1]]
                    res["90(1)"] = ordered_url[extraImgs[2]]
                    res["0(1)"] = ordered_url[extraImgs[3]]
                else:
                    res["360(1)"] = ordered_url[extraImgs[0]]
                    res["270(1)"] = ordered_url[extraImgs[1]]
                    res["180(1)"] = ordered_url[extraImgs[2]]
                    res["90(1)"] = ordered_url[extraImgs[3]]
                    res["0(1)"] = ordered_url[extraImgs[4]]
            elif len(extraImgs) in (2, 3):
                if scan_range > 180:
                    _logger.warning(
                        "incoherent data information to retrieve"
                        "scan extra images angle"
                    )
                elif len(extraImgs) == 3:
                    res["180(1)"] = ordered_url[extraImgs[0]]
                    res["90(1)"] = ordered_url[extraImgs[1]]
                    res["0(1)"] = ordered_url[extraImgs[2]]
                else:
                    res["90(1)"] = ordered_url[extraImgs[0]]
                    res["0(1)"] = ordered_url[extraImgs[1]]
            elif len(extraImgs) == 1:
                res["0(1)"] = ordered_url[extraImgs[0]]
            else:
                raise ValueError(
                    "incoherent data information to retrieve scan" "extra images angle"
                )
        return res

    @property
    def intensity_normalization(self):
        return self._intensity_normalization

    @intensity_normalization.setter
    def intensity_normalization(self, value):
        try:
            method = _IntensityMethod(value)
        except ValueError:
            pass
        else:
            self._intensity_normalization.method = method

    def get_sinogram(
        self,
        line,
        subsampling=1,
        norm_method: str | None = None,
        **kwargs,
    ):
        """
        extract the sinogram from projections

        :param line: which sinogram we want
        :param subsampling: subsampling to apply. Allows to skip some io

        :return: computed sinogram from projections
        """
        if (
            self.projections is not None
            and self.dim_2 is not None
            and line > self.dim_2
        ) or line < 0:
            raise ValueError(f"requested line {line} is not in the scan")
        if self.projections is not None:
            y_dim = ceil(len(self.projections) / subsampling)
            sinogram = numpy.empty((y_dim, self.dim_1))
            _logger.debug(
                f"compute sinogram for line {line} of {self.path} (subsampling: {subsampling})"
            )
            advancement = Progress(
                name=f"compute sinogram for {os.path.basename(self.path)}, line={line},sampling={subsampling}"
            )
            advancement.setMaxAdvancement(len(self.projections))
            projections = self.projections
            o_keys = list(projections.keys())
            o_keys.sort()
            for i_proj, proj_index in enumerate(o_keys):
                if i_proj % subsampling == 0:
                    proj_url = projections[proj_index]
                    proj = silx.io.utils.get_data(proj_url)
                    proj = self.flat_field_correction(
                        projs=[proj], proj_indexes=[proj_index]
                    )[0]
                    sinogram[i_proj // subsampling] = proj[line]
                advancement.increaseAdvancement(1)

            return self._apply_sino_norm(
                sinogram,
                line=line,
                norm_method=norm_method,
                subsampling=subsampling,
                **kwargs,
            )
        else:
            return None

    def _apply_sino_norm(
        self, sinogram, line, norm_method: _IntensityMethod, subsampling=1, **kwargs
    ) -> numpy.ndarray | None:
        if norm_method is not None:
            norm_method = _IntensityMethod(norm_method)
        if norm_method in (None, _IntensityMethod.NONE):
            return sinogram
        elif norm_method is _IntensityMethod.CHEBYSHEV:
            return normalize_chebyshev_2D(sinogram)
        elif norm_method is _IntensityMethod.LSQR_SPLINE:
            return normalize_lsqr_spline_2D(sinogram)
        elif norm_method in (_IntensityMethod.DIVISION, _IntensityMethod.SUBTRACTION):
            # get intensity factor
            if "value" in kwargs:
                intensities = kwargs["value"]
                _logger.info("Apply sinogram normalization from 'value' key")
            elif "dataset_url" in kwargs:
                _logger.info("Apply sinogram normalization from 'dataset_url' key")
                try:
                    if isinstance(kwargs["dataset_url"], DataUrl):
                        url = kwargs["dataset_url"]
                    else:
                        url = DataUrl(path=kwargs["dataset_url"])
                    intensities = get_data(url)
                except Exception as e:
                    _logger.error(f"Fail to load intensities. Error is {e}")
                    return
            else:
                raise KeyError(
                    f"{norm_method.value} requires a value or an url to be computed"
                )
            if intensities is None:
                raise ValueError("provided normalization intensities is None")

            # apply normalization
            if numpy.isscalar(intensities):
                if norm_method is _IntensityMethod.SUBTRACTION:
                    sinogram = sinogram - intensities
                elif norm_method is _IntensityMethod.DIVISION:
                    sinogram = sinogram / intensities
                else:
                    raise NotImplementedError
            elif not isinstance(intensities, numpy.ndarray):
                raise TypeError(
                    f"intensities is expected to be a numpy array not a ({type(intensities)})"
                )
            elif intensities.ndim == 1:
                # in the case intensities is a 1D array: we expect to have one value per projection
                for sl, value in enumerate(intensities):
                    if norm_method is _IntensityMethod.SUBTRACTION:
                        sinogram[sl] = sinogram[sl] - value
                    elif norm_method is _IntensityMethod.DIVISION:
                        sinogram[sl] = sinogram[sl] / value
            elif intensities.ndim in (2, 3):
                # in the case intensities is a 2D array: we expect to have one array per projection (each line has a value)
                # in the case intensities is a 3D array: we expect to have one frame per projection
                for sl, value in enumerate(intensities):
                    if norm_method is _IntensityMethod.SUBTRACTION:
                        sinogram[sl] = sinogram[sl] - value[line]
                    elif norm_method is _IntensityMethod.DIVISION:
                        sinogram[sl] = sinogram[sl] / value[line]
            else:
                raise ValueError(
                    f"{kwargs['dataset_url']} is expected to be 1D, 2D or 3D"
                )
            return sinogram
        else:
            raise ValueError("norm method not handled", norm_method)

    def _frame_flat_field_correction(
        self,
        data: numpy.ndarray | DataUrl,
        dark,
        flat_weights: dict,
        line: None | int = None,
    ):
        """
        compute flat field correction for a provided data from is index
        one dark and two flats (require also indexes)
        """
        assert isinstance(data, (numpy.ndarray, DataUrl))
        if isinstance(data, DataUrl):
            data = get_data(data)
        can_process = True
        if flat_weights in (None, {}):
            if self._notify_ffc_rsc_missing:
                _logger.error(
                    f"cannot make flat field correction, flat not found from {self} or weights not computed"
                )
            can_process = False
        else:
            for flat_index, _ in flat_weights.items():
                if flat_index not in self.reduced_flats:
                    _logger.error(
                        f"flat {flat_index} has been removed, unable to apply flat field"
                    )
                    can_process = False
                elif (
                    self.reduced_flats is not None
                    and self.reduced_flats[flat_index].ndim != 2
                ):
                    _logger.error(
                        "cannot make flat field correction, flat should be of dimension 2"
                    )
                    can_process = False

        if can_process is False:
            self._notify_ffc_rsc_missing = False
            if line is None:
                return data
            else:
                return data[line]

        if len(flat_weights) == 1:
            flat_value = self.reduced_flats[list(flat_weights.keys())[0]]
        elif len(flat_weights) == 2:
            flat_keys = list(flat_weights.keys())
            flat_1 = flat_keys[0]
            flat_2 = flat_keys[1]
            flat_value = (
                self.reduced_flats[flat_1] * flat_weights[flat_1]
                + self.reduced_flats[flat_2] * flat_weights[flat_2]
            )
        else:
            raise ValueError(
                "no more than two flats are expected and"
                "at least one shuold be provided"
            )
        if line is None:
            assert data.ndim == 2
            div = flat_value - dark
            div[div == 0] = 1.0
            return (data - dark) / div
        else:
            assert data.ndim == 1
            div = flat_value[line] - dark[line]
            div[div == 0] = 1
            return (data - dark[line]) / div

    def flat_field_correction(
        self,
        projs: list | tuple | numpy.ndarray,
        proj_indexes: list | tuple | numpy.ndarray | None,
        line: int | None = None,
    ):
        """Apply flat field correction on the given data

        :param projs: list, tuple or numpy.ndarray of projections (numpy array) to apply correction on
        :param proj_indexes: list, tuple or numpy.ndarray of indexes of the projection in
                             the acquisition sequence. Provide None to use the index in the middle
                             of the flats taken.
        :param line: index of the line to apply flat filed. If not provided
                     consider we want to apply flat filed on the entire frame
        :return: corrected data: list of numpy array
        """
        assert isinstance(projs, (list, tuple, numpy.ndarray))
        assert proj_indexes is None or isinstance(
            proj_indexes, (list, tuple, numpy.ndarray)
        )
        assert isinstance(line, (type(None), int))

        def has_missing_keys():
            if proj_indexes is None:
                return False
            for proj_index in proj_indexes:
                if proj_index not in self._flats_weights:
                    return False
            return True

        def return_without_correction():
            def load_data(proj):
                if isinstance(proj, DataUrl):
                    return get_data(proj)
                else:
                    return proj

            if line is not None:
                res = [
                    load_data(proj)[line] if isinstance(proj, DataUrl) else proj
                    for proj in projs
                ]
            else:
                res = [
                    load_data(proj) if isinstance(proj, DataUrl) else proj
                    for proj in projs
                ]

            return res

        if self._flats_weights in (None, {}) or has_missing_keys():
            self._flats_weights = self._get_flats_weights()

        if self._flats_weights in (None, {}):
            if self._notify_ffc_rsc_missing:
                _logger.error("Unable to compute flat weights")
                self._notify_ffc_rsc_missing = False
            return return_without_correction()

        darks = self._reduced_darks
        if darks is not None and len(darks) > 0:
            # take only one dark into account for now
            dark = list(darks.values())[0]
        else:
            dark = None

        if dark is None:
            if self._notify_ffc_rsc_missing:
                _logger.error("cannot make flat field correction, dark not found")
                self._notify_ffc_rsc_missing = False
            return return_without_correction()

        if dark is not None and dark.ndim != 2:
            if self._notify_ffc_rsc_missing:
                _logger.error(
                    "cannot make flat field correction, dark should be of "
                    "dimension 2"
                )
                self._notify_ffc_rsc_missing = False
                return return_without_correction()

        return numpy.array(
            [
                self._frame_flat_field_correction(
                    data=frame,
                    dark=dark,
                    flat_weights=(
                        self._flats_weights[proj_i]
                        if proj_i in self._flats_weights
                        else None
                    ),
                    line=line,
                )
                for frame, proj_i in zip(projs, proj_indexes)
            ]
        )

    def _get_flats_weights(self):
        """compute flats indexes to use and weights for each projection"""
        if self.reduced_flats is None:
            return None
        flats_indexes = sorted(self.reduced_flats.keys())

        def get_weights(proj_index):
            if proj_index in flats_indexes:
                return {proj_index: 1.0}
            pos = bisect_left(flats_indexes, proj_index)
            left_pos = flats_indexes[pos - 1]
            if pos == 0:
                return {flats_indexes[0]: 1.0}
            elif pos > len(flats_indexes) - 1:
                return {flats_indexes[-1]: 1.0}
            else:
                right_pos = flats_indexes[pos]
                delta = right_pos - left_pos
                return {
                    left_pos: 1 - (proj_index - left_pos) / delta,
                    right_pos: 1 - (right_pos - proj_index) / delta,
                }

        if self.reduced_flats is None or len(self.reduced_flats) == 0:
            return {}
        else:
            res = {}
            for proj_index in self.projections:
                res[proj_index] = get_weights(proj_index=proj_index)
            return res

    def get_projections_intensity_monitor(self):
        """return intensity monitor values for projections"""
        raise NotImplementedError("Base class")

    def get_flat_expected_location(self):
        raise NotImplementedError("Base class")

    def get_dark_expected_location(self):
        raise NotImplementedError("Base class")

    def get_projection_expected_location(self):
        raise NotImplementedError("Base class")

    def get_energy_expected_location(self):
        raise NotImplementedError("Base class")

    def get_sample_detector_distance_expected_location(self):
        raise NotImplementedError("Base class")

    def get_pixel_size_expected_location(self):
        raise NotImplementedError("Base class")

    def get_relative_file(self, file_name: str, with_dataset_prefix=True) -> str | None:
        """
        :param file_name: name of the file to create
        :param with_dataset_prefix: If True will prefix the requested file by the dataset name like datasetname_file_name

        :return: path to the requested file according to the 'Scan' / 'dataset' location. Return none if Scan has no path
        """
        raise NotImplementedError("Base class")

    def get_dataset_basename(self) -> str:
        raise NotImplementedError("Base class")

    def _dump_frame_dict(
        self,
        frames: dict[int, numpy.ndarray],
        output_urls,
        frames_metadata: ReducedFramesInfos | None,
        metadata_output_urls: tuple | None,
        overwrite: bool = False,
    ):
        """
        utils function to save some frames in set of output_urls

        Behavior with HDF5: it expects to have a dedicated group where it can save the different frame with indices.
        It will do a first iteration at this group level to remove unused dataset and will overwrite the one he can in order to reduced memory print
        """
        saver = ReduceFrameSaver(
            frames=frames,
            output_urls=[self._format_url(url) for url in output_urls],
            frames_metadata=frames_metadata,
            metadata_output_urls=[
                self._format_url(url) for url in metadata_output_urls
            ],
            overwrite=overwrite,
        )
        saver.save()

    def _format_url(self, url: DataUrl):
        """
        Format an url according to the scan metadata. Solves ''entry' and 'scan_prefix' keywords
        """
        assert isinstance(
            url, DataUrl
        ), f"url should be an instance of {DataUrl}. Got {type(url)}"
        if hasattr(self, "entry"):
            entry = self.entry
        else:
            entry = "entry"

        def format_data_path(data_path) -> str:
            """Format only 'entry' field. This is up to the 'ReduceFrameSaver' to take care of 'index' and 'index_zfill4'"""
            if data_path is not None:
                data_path = data_path.format(
                    entry=entry, index="{index}", index_zfill4="{index_zfill4}"
                )
            return data_path

        def format_file_path(file_path) -> str:
            """Format only 'entry' and 'scan_prefix' field. This is up to the 'ReduceFrameSaver' to take care of 'index' and 'index_zfill4'"""
            if file_path is not None:
                if not os.path.isabs(file_path):
                    if self.path is None:
                        raise ValueError("No dataset path provided")

                    file_path = os.path.join(self.path, file_path)

                file_path = file_path.format(
                    entry=entry,
                    scan_prefix=self.get_dataset_basename(),
                    index="{index}",
                    index_zfill4="{index_zfill4}",
                )
            return file_path

        return DataUrl(
            file_path=format_file_path(url.file_path()),
            data_path=format_data_path(url.data_path()),
            scheme=url.scheme(),
            data_slice=url.data_slice(),
        )

    def _load_frame_dict(
        self, inputs_urls, metadata_input_urls, return_as_url=False
    ) -> dict:
        """
        :note: note on pattern:
            * handled patterns are:
                * file_path pattern:
                    * {index}: only handled for edf files
                    * {index_zfill4}: only handled for edf files
                    * Only one usage of index and index_zfill4 can be done. Having several {index} or one {index} and one {index_zfill4} will fail
                * data_path pattern:
                    * {entry}
                    * {index}: works only if set at the end of the path (as dataset name)
                    * {index_zfill4}: works only if set at the end of the path (as dataset name)
        :return: tuple(frames_data, frames_metadata).
                 * frames_data: dict with frame index in the acquisition sequence as key. Value is the frame as a numpy array if return_as_url is False else this is a DataUrl to the frame
                 * frames_metadata: return an instance of ReducedFramesInfos. We consider this is too small to use the DataUrl mecanism when return_as_url set to True
        """
        from tomoscan.esrf.scan.utils import (
            get_files_from_pattern,
        )  # avoid cyclic import

        if self.path is None:
            raise ValueError("No dataset path provided")

        res_data = {}
        entry = "entry"
        if hasattr(self, "entry"):
            entry = self.entry

        res_metadata = ReducedFramesInfos()

        # load frames infos
        for url in inputs_urls:
            data_path = url.data_path()
            if data_path is not None:
                data_path = data_path.format(
                    entry=entry, index_zfill4="{index_zfill4}", index="{index}"
                )
                # we don't want to handle index_zfill4 and index at this level
            file_pattern = url.file_path()
            if file_pattern is not None:
                file_pattern = file_pattern.format(
                    entry,
                    entry,
                    index_zfill4="{index_zfill4}",
                    index="{index}",
                    scan_prefix=self.get_dataset_basename(),
                )
                # we don't want to handle index_zfill4 and index at this level
            scheme = url.scheme()

            frames_path_and_index = []
            # list of tuples (frame_file, frame_index). frame_index can be None if not found
            patterns = ("index_zfill4", "index")
            contains_patterns = False
            for pattern in patterns:
                if pattern in file_pattern:
                    contains_patterns = True
                    files_from_pattern = get_files_from_pattern(
                        file_pattern=file_pattern,
                        pattern=pattern,
                        research_dir=self.path,
                    )
                    for frame_index, frame_file in files_from_pattern.items():
                        frames_path_and_index.append(
                            (os.path.join(self.path, frame_file), frame_index)
                        )
            if not contains_patterns:
                frames_path_and_index.append(
                    (os.path.join(self.path, file_pattern), None)
                )

            def format_data_path(data_path):
                index_zfill4_pattern = False
                if data_path.endswith("{index_zfill4}"):
                    index_zfill4_pattern = True
                    data_path = data_path[: -len("{index_zfill4}")]
                if data_path.endswith("{index}"):
                    data_path = data_path[: -len("{index}")]
                if data_path.endswith("/"):
                    data_path = data_path[:-1]
                return data_path, index_zfill4_pattern

            for frame_file_path, frame_index in frames_path_and_index:
                if scheme == "fabio":
                    if not os.path.exists(frame_file_path):
                        continue
                    try:
                        handler = fabio.open(frame_file_path)
                        with fabio.open(frame_file_path) as handler:
                            if handler.nframes > 1:
                                _logger.warning(
                                    f"{frame_file_path} is expected to have one frame. Only the first one will be picked"
                                )
                            if frame_index in res_data:
                                _logger.error(
                                    f"two frames found with the same index {frame_index}"
                                )
                            if return_as_url:
                                res_data[frame_index] = DataUrl(
                                    file_path=frame_file_path, scheme="fabio"
                                )
                            else:
                                res_data[frame_index] = handler.data
                            if "SRCUR" in handler.header:
                                res_metadata.machine_current.append(
                                    float(handler.header["SRCUR"])
                                )
                            if "CountTime" in handler.header:
                                res_metadata.count_time.append(
                                    float(handler.header["CountTime"])
                                )

                    except OSError as e:
                        _logger.error(e)
                elif scheme in ("hdf5", "silx"):
                    data_path, index_zfill4_pattern = format_data_path(data_path)

                    if not os.path.exists(frame_file_path):
                        continue
                    with hdf5_open(frame_file_path) as h5s:
                        dataset_or_group = h5s[data_path]
                        if isinstance(dataset_or_group, h5py.Dataset):
                            idx = None
                            if dataset_or_group.name.isnumeric():
                                try:
                                    idx = int(dataset_or_group)
                                except ValueError:
                                    idx = None
                            if return_as_url:
                                res_data[idx] = DataUrl(
                                    file_path=frame_file_path,
                                    data_path=data_path,
                                    scheme="silx",
                                )
                            else:
                                res_data[idx] = dataset_or_group[()]
                        else:
                            assert isinstance(
                                dataset_or_group, h5py.Group
                            ), f"expect a group not {type(dataset_or_group)}"
                            # browse children
                            for name, item in dataset_or_group.items():
                                if isinstance(item, h5py.Dataset):
                                    if name.isnumeric():
                                        if index_zfill4_pattern and not len(name) == 4:
                                            continue
                                        else:
                                            try:
                                                idx = int(name)
                                            except ValueError:
                                                _logger.info(
                                                    f"fail to cast {name} as a integer"
                                                )
                                                continue
                                            if return_as_url:
                                                res_data[idx] = DataUrl(
                                                    file_path=frame_file_path,
                                                    data_path=data_path + "/" + name,
                                                    scheme="silx",
                                                )
                                            else:
                                                res_data[idx] = dataset_or_group[name][
                                                    ()
                                                ]
                else:
                    raise ValueError(
                        f"scheme {scheme} is not handled. Should be fabio, silx of hdf5"
                    )

        def get_unit(attrs) -> pint.Unit | None:
            if "unit" in attrs:
                unit = attrs["unit"]
            elif "units":
                unit = attrs["units"]
            if unit == "kev":
                unit = _ureg.keV
            if unit is not None:
                return pint.Unit(unit)

            return None

        # load frames metadata
        if metadata_input_urls is not None:
            for url in metadata_input_urls:
                metadata_file = url.file_path()
                metadata_file = metadata_file.format(
                    scan_prefix=self.get_dataset_basename(),
                )
                if not os.path.isabs(metadata_file):
                    metadata_file = os.path.join(self.path, metadata_file)
                data_path = url.data_path().format(
                    entry=entry,
                )

                if scheme in ("hdf5", "silx"):
                    if not os.path.exists(metadata_file):
                        continue
                    with hdf5_open(metadata_file) as h5s:
                        if data_path not in h5s:
                            continue
                        parent_group = h5s[data_path]
                        if ReducedFramesInfos.COUNT_TIME_KEY in parent_group:
                            count_time = silx.io.utils.h5py_read_dataset(
                                parent_group[ReducedFramesInfos.COUNT_TIME_KEY]
                            )
                            unit = get_unit(
                                attrs=parent_group[
                                    ReducedFramesInfos.COUNT_TIME_KEY
                                ].attrs,
                            )
                            if unit is None:
                                unit = _ureg.second
                            res_metadata.count_time = (
                                (count_time * unit).to(_ureg.second).magnitude
                            )

                        if ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY in parent_group:
                            machine_current = silx.io.utils.h5py_read_dataset(
                                parent_group[
                                    ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY
                                ]
                            )

                            unit = get_unit(
                                attrs=parent_group[
                                    ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY
                                ].attrs,
                            )
                            if unit is None:
                                unit = _ureg.ampere
                            res_metadata.machine_current = (
                                (machine_current * unit).to(_ureg.ampere).magnitude
                            )

        return res_data, res_metadata

    def save_reduced_darks(
        self,
        darks: dict,
        output_urls: tuple,
        darks_infos: ReducedFramesInfos | None = None,
        metadata_output_urls: tuple | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Dump computed dark (median / mean...) into files.

        :param darks: dictionary with frame indices as key (int) and a 2D numpy array as value.
        :param output_urls: tuple of silx DataUrl, where to save the darks. Default value is usually provided by children class directly.
                            You better know what you are doing if you modify the default value.
        :param darks_infos: information regarding darks (metadata) like the machine current, exposure time...
        :param metadata_output_urls: tuple of silx DataUrl, where to save the metadata / darks information.  Default value is usually provided by children class directly
                            You better know what you are doing if you modify the default value.
        :param overwrtie: if the output files exist then will overwrite them.

        Here is an example on how to save your own reduced dark / flat

        .. code-block:: python

            from tomoscan.framereducer.reducedframesinfos import ReducedFramesInfos
            ...

            scan = ... # scan must be an instance of TomoScanBase like NXtomoScan()
            darks_infos.count_time = [2.5]
            darks_infos.machine_current = [13.1]
            scan.save_reduced_darks(
                darks={
                    0: dark_frame,  # dark_frame is a 2d numpy array
                },
                darks_infos=darks_infos,
                overwrite=True,
            )
        """
        self._dump_frame_dict(
            frames=darks,
            output_urls=output_urls,
            frames_metadata=darks_infos,
            metadata_output_urls=metadata_output_urls,
            overwrite=overwrite,
        )

    def load_reduced_darks(
        self,
        inputs_urls: tuple,
        metadata_input_urls=None,
        return_as_url: bool = False,
        return_info: bool = False,
    ) -> dict | tuple:
        """
        load reduced dark (median / mean...) into files

        :param inputs_urls: where to load the reduced darks. A default value is provided by the children class.
                            You better know what you are doing if you modify the default value.
        :param metadata_input_urls: where to load the reduced darks metadata. A default value is provided by the children class.
                                    You better know what you are doing if you modify the default value.
        :param return_as_url: if True then instead of returning the reduced frames as 2D numpy arrays it will return them as a silx DataUrl
        :param return_info: if False only return return the dict of reduced frames (frame index as key (int) and frame as a 2D numpy array or silx Data Url as value)
                            if True then return (dict of reduced frames, darks info / metadata)

        Here is an example of usage:

        .. code-block:: python

            scan = ... # scan must be an instance of TomoScanBase like NXtomoScan()
            reduced_darks = scan.load_reduced_darks()
            reduced_darks, darks_infos = scan.load_reduced_darks(return_info=True)

            dark_frame_np_array = reduced_darks[0]
        """
        darks, infos = self._load_frame_dict(
            inputs_urls=inputs_urls,
            return_as_url=return_as_url,
            metadata_input_urls=metadata_input_urls,
        )
        if return_info:
            return darks, infos
        else:
            return darks

    def save_reduced_flats(
        self,
        flats: dict,
        output_urls: tuple,
        flats_infos: ReducedFramesInfos | None = None,
        metadata_output_urls: tuple | None = None,
        overwrite: bool = False,
    ) -> None:
        """
        Dump computed dark (median / mean...) into files.

        :param flats: dictionary with frame indices as key (int) and a 2D numpy array as value.
        :param output_urls: tuple of silx DataUrl, where to save the flats. Default value is usually provided by children class directly.
                            You better know what you are doing if you modify the default value.
        :param flats_infos: information regarding flats (metadata) like the machine current, exposure time...
        :param metadata_output_urls: tuple of silx DataUrl, where to save the metadata / flats information.  Default value is usually provided by children class directly
                            You better know what you are doing if you modify the default value.
        :param overwrite: if the output files exist then will overwrite them.

        Here is an example on how to save your own reduced dark / flat

        .. code-block:: python

            from tomoscan.framereducer.reducedframesinfos import ReducedFramesInfos
            ...

            scan = ... # scan must be an instance of TomoScanBase like NXtomoScan()
            flats_infos = ReducedFramesInfos()
            flats_infos.count_time = [2.5, 1.2]
            flats_infos.machine_current = [12.5, 13.1]
            # for normalization the first reduced flat (at index 1) will have 2.5 as count time and 12.5 as machine current
            # the second reduced flat frame (at index 1002) will have 1.2 as count time and 13.1 as machine current
            scan.save_reduced_darks(
                darks={
                    1: flat_frame_1,     # flat_frame_1 is a 2d numpy array
                    1002: flat_frame_2,  # flat_frame_2 is another 2d numpy array
                },
                flats_infos=flats_infos,
                overwrite=True,
            )
        """
        self._dump_frame_dict(
            frames=flats,
            output_urls=output_urls,
            frames_metadata=flats_infos,
            metadata_output_urls=metadata_output_urls,
            overwrite=overwrite,
        )

    def load_reduced_flats(
        self,
        inputs_urls: tuple,
        metadata_input_urls=None,
        return_as_url: bool = False,
        return_info: bool = False,
    ) -> dict | tuple:
        """
        load reduced flats frames

        :param inputs_urls: where to load the reduced flats. A default value is provided by the children class.
                            You better know what you are doing if you modify the default value.
        :param metadata_input_urls: where to load the reduced flats metadata. A default value is provided by the children class.
                                    You better know what you are doing if you modify the default value.
        :param return_as_url: if True then instead of returning the reduced frames as 2D numpy arrays it will return them as a silx DataUrl
        :param return_info: if False only return return the dict of reduced frames (frame index as key (int) and frame as a 2D numpy array or silx Data Url as value)
                            if True then return (dict of reduced frames, flats info / metadata)

        Here is an example of usage:

        .. code-block:: python

            scan = ... # scan must be an instance of TomoScanBase like NXtomoScan()
            reduced_flats = scan.load_reduced_flats()
            reduced_flats, flats_infos = scan.load_reduced_flats(return_info=True)

        """
        flats, infos = self._load_frame_dict(
            inputs_urls=inputs_urls,
            return_as_url=return_as_url,
            metadata_input_urls=metadata_input_urls,
        )
        if return_info:
            return flats, infos
        else:
            return flats

    def compute_reduced_flats(
        self,
        reduced_method="median",
        overwrite=True,
        output_dtype=None,
        return_info=False,
    ):
        """
        :param ReduceMethod method: method to compute the flats
        :param overwrite: if some flats have already been computed will overwrite them
        :param return_info: do we return (reduced_frames, info) or directly reduced_frames
        """
        if self.FRAME_REDUCER_CLASS is None:
            raise ValueError("no frame reducer class provided")
        frame_reducer = self.FRAME_REDUCER_CLASS(  # pylint: disable=E1102
            scan=self,
            reduced_method=reduced_method,
            target="flats",
            overwrite=overwrite,
            output_dtype=output_dtype,
        )
        reduced_frames, infos = frame_reducer.run()
        if return_info:
            return reduced_frames, infos
        else:
            return reduced_frames

    def compute_reduced_darks(
        self,
        reduced_method="mean",
        overwrite=True,
        output_dtype=None,
        return_info=False,
    ):
        """
        :param ReduceMethod method: method to compute the flats
        :param overwrite: if some darks have already been computed will overwrite them
        :param return_info: do we return (reduced_frames, info) or directly reduced_frames
        """
        if self.FRAME_REDUCER_CLASS is None:
            raise ValueError("no frame reducer class provided")
        frame_reducer = self.FRAME_REDUCER_CLASS(  # pylint: disable=E1102
            scan=self,
            reduced_method=reduced_method,
            target="darks",
            overwrite=overwrite,
            output_dtype=output_dtype,
        )
        reduced_frames, infos = frame_reducer.run()
        if return_info:
            return reduced_frames, infos
        else:
            return reduced_frames

    @staticmethod
    def get_volume_output_file_name(z=None, suffix=None):
        """if used by tomwer and nabu this should help for tomwer to find out the output files of anbu from a configuration file. Could help to get some normalization there"""
        raise NotImplementedError


def _find_closest_index(values, query):
    pos = bisect_left(values, query)
    if pos == 0:
        return 0
    if pos >= len(values):
        return len(values) - 1
    return pos - 1 if abs(values[pos - 1] - query) < abs(values[pos] - query) else pos


def projections_angles_to_indices(
    angles: list, projections_indices: list, rotation_angles: list
):
    """
    From a list of angles (in degrees), return the corresponding indices.

    :param angles: Rotation angles
    :param projections_indices: List of indices corresponding to each projection
    :param rotation_angles: List of all rotation angles of the projections
    """
    ignore_projections_indices = []
    for angle in angles:
        proj_idx0 = _find_closest_index(rotation_angles, angle)
        proj_idx = projections_indices[proj_idx0]
        ignore_projections_indices.append(proj_idx)
    return ignore_projections_indices


def projection_angular_range_to_indices(
    angular_range: tuple, projections_indices: list, rotation_angles: list
):
    """
    From an angular range, return the corresponding indices of projections.

    :param angular_range: Tuple defining the angular range [min, max] where boundaries are included.
    :param projections_indices: List of indices corresponding to each projection
    :param rotation_angles: List of all rotation angles of the projections
    """
    ignore_projections_indices = []
    angle_min, angle_max = angular_range
    projections_indices = numpy.array(projections_indices)
    for proj_idx, angle in zip(projections_indices, rotation_angles):
        if angle_min <= angle and angle <= angle_max:
            ignore_projections_indices.append(proj_idx)
    return ignore_projections_indices



"""General utils"""

from .decorator import docstring  # noqa F401
from .geometry import BoundingBox1D, BoundingBox3D, get_subvolume_shape  # noqa F401
import numpy


def is_numpy_scalar_dtype(par):
    try:
        numpy.dtype(par)
        res = True
    except TypeError:
        res = False
    return res



# coding: utf-8

import os
import shutil
import tempfile
import unittest

from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.esrf.mock import MockEDF, MockNXtomo


class TestMockEDFScan(unittest.TestCase):
    """Test that mock scan are adapted to other unit test"""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.scan_id = os.path.join(self.tmpdir, "myscan")

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def testScanEvolution360(self):
        """Test get scan evolution from a mock scan with a 360 range and no
        extra radio"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=5,
            nRecons=1,
            nPagRecons=1,
            dim=10,
            scan_range=360,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 5)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0002.edf"),
        )
        self.assertTrue(360 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[360].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0004.edf"),
        )

    def testScanEvolution180(self):
        """Test get scan evolution from a mock scan with a 180 range and no
        extra radio"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=8,
            nRecons=0,
            nPagRecons=0,
            dim=10,
            scan_range=180,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 8)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0007.edf"),
        )
        self.assertFalse(360 in scan_dynamic)

    def testScanEvolution360Extra4(self):
        """Test get scan evolution from a mock scan with a 360 range and 4
        extra radio"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=21,
            nRecons=0,
            nPagRecons=0,
            dim=10,
            scan_range=360,
            n_extra_radio=4,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 21 + 4)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0010.edf"),
        )
        self.assertTrue(360 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[360].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0020.edf"),
        )
        extra_angles = (0, 90, 180, 270)  # note: 0(1) is the last file acquire
        for iAngle, angle in enumerate(extra_angles):
            angle_id = str(angle) + "(1)"
            self.assertTrue(angle_id in scan_dynamic)
            file_name = os.path.join(
                self.tmpdir, "myscan", "myscan_%04d.edf" % (21 + 4 - 1 - iAngle)
            )
            self.assertEqual(scan_dynamic[angle_id].file_path(), file_name)

    def testScanEvolution180Extra2(self):
        """Test get scan evolution from a mock scan with a 360 range and 3
        extra radios"""
        scan = MockEDF.mockScan(
            scanID=self.scan_id,
            nRadio=4,
            nRecons=2,
            nPagRecons=2,
            dim=10,
            scan_range=180,
            n_extra_radio=2,
        )
        scan_dynamic = scan.get_proj_angle_url()
        self.assertEqual(len(scan_dynamic), 4 + 2)
        self.assertTrue(0 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[0].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0000.edf"),
        )
        self.assertTrue(180 in scan_dynamic)
        self.assertEqual(
            scan_dynamic[180].file_path(),
            os.path.join(self.tmpdir, "myscan", "myscan_0003.edf"),
        )
        self.assertTrue(360 not in scan_dynamic)
        extra_angles = (0, 90)  # note: 0(1) is the last file acquire
        for iAngle, angle in enumerate(extra_angles):
            angle_id = str(angle) + "(1)"
            self.assertTrue(angle_id in scan_dynamic)
            file_name = os.path.join(
                self.tmpdir, "myscan", "myscan_%04d.edf" % (4 + 2 - 1 - iAngle)
            )
            self.assertEqual(scan_dynamic[angle_id].file_path(), file_name)


class TestMockNXtomo(unittest.TestCase):
    """Test the MockNXtomo file"""

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def test_creation(self):
        mock = MockNXtomo(scan_path=self.folder, n_proj=10, n_ini_proj=10)
        self.assertEqual(
            mock.scan_master_file,
            os.path.join(self.folder, os.path.basename(self.folder) + ".h5"),
        )
        tomoScan = NXtomoScan(mock.scan_path, entry=mock.scan_entry)
        self.assertEqual(len(NXtomoScan.get_valid_entries(mock.scan_master_file)), 1)
        tomoScan.update()
        self.assertEqual(tomoScan.scan_range, 360)
        self.assertEqual(len(tomoScan.projections), 10)

    def testSimpleMockCreationOneCall(self):
        """Test mock of an acquisition starting by one dark, then 10 ref,
        then 20 radios, then 10 'final' ref and 2 alignment radio"""
        mock = MockNXtomo(
            scan_path=self.folder,
            n_proj=20,
            n_ini_proj=20,
            n_alignement_proj=2,
            create_ini_dark=True,
            create_ini_ref=True,
            create_final_ref=True,
            n_refs=10,
        )
        self.assertTrue(0 in mock.scan.darks.keys())
        self.assertTrue(1 in mock.scan.flats.keys())
        self.assertTrue(10 in mock.scan.flats.keys())
        self.assertEqual(mock.scan.intensity_monitor, None)
        self.assertEqual(mock.scan.frames[0].intensity_monitor, None)

    def testMockIntensityMonitor(self):
        """Test intensity monitor is correctly read"""
        mock = MockNXtomo(
            scan_path=self.folder,
            n_proj=20,
            n_ini_proj=20,
            n_alignement_proj=2,
            create_ini_dark=True,
            create_ini_ref=True,
            create_final_ref=True,
            n_refs=10,
            intensity_monitor=True,
        )
        self.assertEqual(len(mock.scan.intensity_monitor), len(mock.scan.frames))
        self.assertNotEqual(mock.scan.frames[0].intensity_monitor, None)
